"""Code for Symbolic Fourier Approximation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
from .dft import DiscreteFourierTransform
from .mcb import MultipleCoefficientBinning
from ..base import UnivariateTransformerMixin


class SymbolicFourierApproximation(BaseEstimator, UnivariateTransformerMixin):
    """Symbolic Fourier Approximation.

    Parameters
    ----------
    n_coefs : None, int or float (default = None)
        The number of Fourier coefficients to keep. If None, all the Fourier
        coeeficients are kept. If an integer, the ``n_coefs`` most significant
        Fourier coefficients are returned if ``anova=True``, otherwise the
        first ``n_coefs`` Fourier coefficients are returned. If a float, it
        represents a percentage of the size of each time series and must be
        between 0 and 1. The number of coefficients will be computed as
        ``ceil(n_coefs * (n_timestamps - 1))`` if ``drop_sum=True`` and
        ``ceil(n_coefs * n_timestamps)`` if ``drop_sum=False``.

    n_bins : int (default = 4)
        The number of bins to produce. The intervals for the bins are
        determined by the minimum and maximum of the input data. It must
        be between 2 and 26.

    strategy : str (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    drop_sum : bool (default = False)
        If True, the first Fourier coefficient (i.e. the sum of the time
        series) is dropped. If False, the real part of the first Fourier
        coefficient is kept.

    anova : bool (default = False)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    norm_mean : bool (default = False)
        If True, center the data before scaling. If ``norm_mean=True`` and
        ``anova=False``, the first Fourier coefficient will be dropped.

    norm_std : bool (default = False)
        If True, scale the data to unit variance.

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used if `n_bins` is lower than 27, otherwise the alphabet
        will be defined to [chr(i) for i in range(n_bins)]. If 'ordinal',
        integers are used.

    Attributes
    ----------
    bin_edges_ : array, shape = (n_bins - 1,) or (n_timestamps, n_bins - 1)
        Bin edges with shape = (n_bins - 1,) if ``strategy='normal'`` or
        (n_timestamps, n_bins - 1) otherwise.

    support_ : array, shape = (n_coefs,)
        Indices of the kept Fourier coefficients.

    References
    ----------
    .. [1] P. Schäfer, and M. Högqvist, "SFA: A Symbolic Fourier Approximation
           and Index for Similarity Search in High Dimensional Datasets",
           International Conference on Extending Database Technology,
           15, 516-527 (2012).

    Examples
    --------
    >>> from pyts.approximation import SymbolicFourierApproximation
    >>> from pyts.datasets import load_gunpoint
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = SymbolicFourierApproximation(n_coefs=4)
    >>> X_new = transformer.fit_transform(X)
    >>> X_new.shape
    (50, 4)

    """

    def __init__(self, n_coefs=None, n_bins=4, strategy='quantile',
                 drop_sum=False, anova=False, norm_mean=False, norm_std=False,
                 alphabet=None):
        self.n_coefs = n_coefs
        self.drop_sum = drop_sum
        self.anova = anova
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.n_bins = n_bins
        self.strategy = strategy
        self.alphabet = alphabet

    def fit(self, X, y=None):
        """Select Fourier coefficients and compute bin edges for each feature.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        y : None or array-like, shape = (n_samples,) (default = None)
            Class labels for each sample. Only used if ``anova=True`` or
            ``strategy='entropy'.``

        """
        dft = DiscreteFourierTransform(
            n_coefs=self.n_coefs, drop_sum=self.drop_sum, anova=self.anova,
            norm_mean=self.norm_mean, norm_std=self.norm_std
        )
        mcb = MultipleCoefficientBinning(
            n_bins=self.n_bins, strategy=self.strategy, alphabet=self.alphabet
        )
        self._pipeline = Pipeline([('dft', dft), ('mcb', mcb)])
        self._pipeline.fit(X, y)
        self.support_ = self._pipeline.named_steps['dft'].support_
        self.bin_edges_ = self._pipeline.named_steps['mcb'].bin_edges_
        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array, shape = (n_samples, n_coefs)
            Transformed data.

        """
        check_is_fitted(self, ['support_', 'bin_edges_'])
        return self._pipeline.transform(X)

    def fit_transform(self, X, y=None):
        """Fit then transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        y : None or array-like, shape = (n_samples,)
            Class labels for each sample. Only used if ``anova=True`` or
            ``strategy='entropy'.``

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_coefs)
            Transformed data.

        """
        dft = DiscreteFourierTransform(
            n_coefs=self.n_coefs, drop_sum=self.drop_sum, anova=self.anova,
            norm_mean=self.norm_mean, norm_std=self.norm_std
        )
        mcb = MultipleCoefficientBinning(
            n_bins=self.n_bins, strategy=self.strategy, alphabet=self.alphabet
        )
        self._pipeline = Pipeline([('dft', dft), ('mcb', mcb)])
        X_sfa = self._pipeline.fit_transform(X, y)
        self.support_ = self._pipeline.named_steps['dft'].support_
        self.bin_edges_ = self._pipeline.named_steps['mcb'].bin_edges_
        return X_sfa



"""Testing for Piecewise Aggregate Approximation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.approximation.paa import _paa
from pyts.approximation import PiecewiseAggregateApproximation


X = [np.arange(6), [0, 5, 1, 2, 4, 3]]


@pytest.mark.parametrize(
    'X, start, end, arr_desired',
    [([[0, 1, 2, 3], [1, 3, 0, 2], [4, 8, 2, 6]], [0, 2], [2, 4],
      [[0.5, 2.5], [2.0, 1.0], [6.0, 4.0]]),

     ([[0, 1, 2, 3], [1, 3, 0, 2], [4, 8, 2, 6]], [0, 1, 2], [2, 3, 4],
      [[0.5, 1.5, 2.5], [2.0, 1.5, 1.0], [6.0, 5.0, 4.0]]),

     ([[0, 1, 2, 3], [1, 3, 0, 2], [4, 8, 2, 6]], [0, 1, 2, 3], [1, 2, 3, 4],
      [[0, 1, 2, 3], [1, 3, 0, 2], [4, 8, 2, 6]])]
)
def test_actual_results_paa(X, start, end, arr_desired):
    """Test that the actual results are the expected ones."""
    X = np.asarray(X)
    start = np.asarray(start)
    end = np.asarray(end)
    n_timestamps_new = start.size
    n_samples, n_timestamps = X.shape
    arr_actual = _paa(X, n_samples, n_timestamps, start, end, n_timestamps_new)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'window_size': None, 'output_size': None}, TypeError,
      "'window_size' and 'output_size' cannot be both None."),

     ({'window_size': '3'}, TypeError,
      "If specified, 'window_size' must be an integer or a float."),

     ({'window_size': None, 'output_size': '3'}, TypeError,
      "If specified, 'output_size' must be an integer or a float."),

     ({'window_size': 0}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'window_size': 2.}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(2.)),

     ({'window_size': None, 'output_size': 0}, ValueError,
      "If 'output_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'window_size': None, 'output_size': 2.}, ValueError,
      "If 'output_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(2.))]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    paa = PiecewiseAggregateApproximation(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        paa.transform(X)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({}, X),
     ({'window_size': 2}, [[0.5, 2.5, 4.5], [2.5, 1.5, 3.5]]),

     ({'window_size': None, 'output_size': 3},
      [[0.5, 2.5, 4.5], [2.5, 1.5, 3.5]]),

     ({'window_size': None, 'output_size': 0.5},
      [[0.5, 2.5, 4.5], [2.5, 1.5, 3.5]]),

     ({'window_size': 0.5}, [[1, 4], [2, 3]])]
)
def test_actual_results(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = PiecewiseAggregateApproximation(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



import pytest

from pyts.utils.deprecation import _is_deprecated
from pyts.utils.deprecation import deprecated

# Author: <hicham.janati@inria.fr>
# Adapted from sklearn.utils.tests.test_deprecation


@deprecated('qwerty')
class MockClass1:
    pass


class MockClass2:
    @deprecated('mockclass2_method')
    def method(self):
        pass


class MockClass3:
    @deprecated()
    def __init__(self):
        pass


class MockClass4:
    pass


@deprecated()
def mock_function():
    return 10


def test_deprecated():
    with pytest.warns(DeprecationWarning, match="qwerty"):
        MockClass1()

    with pytest.warns(DeprecationWarning, match="mockclass2_method"):
        MockClass2().method()

    with pytest.warns(DeprecationWarning, match="deprecated"):
        MockClass3()

    with pytest.warns(DeprecationWarning):
        val = mock_function()
        assert val == 10


def test_is_deprecated():
    # Test if _is_deprecated helper identifies wrapping via deprecated
    # NOTE it works only for class methods and functions
    assert _is_deprecated(MockClass1.__init__)
    assert _is_deprecated(MockClass2().method)
    assert _is_deprecated(MockClass3.__init__)
    assert not _is_deprecated(MockClass4.__init__)
    assert _is_deprecated(mock_function)



"""Functions to make datasets."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from sklearn.utils import check_array, check_random_state
from sklearn.utils import shuffle as utils_shuffle


def make_cylinder_bell_funnel(
    n_samples=30, weights=None, shuffle=True, random_state=None,
    return_params=False
):
    r"""Make a Cylinder-Bell-Funnel dataset.

    The classes are coded with the following meaning:

        - 0: Cylinder
        - 1: Bell
        - 2: Funnel

    Parameters
    ----------
    n_samples : int (default = 30)
        The number of time series.

    weights : list of floats or None (default=None)
        The proportions of samples assigned to each class. If None, then
        classes are balanced. Note that if ``len(weights) == 2``,
        then the last class weight is automatically inferred.
        More than ``n_samples`` samples may be returned if the sum of
        ``weights`` exceeds 1.

    shuffle : bool (default = True)
        If True, shuffle the samples.

    return_params : bool (default = False)
        If True, a dictionary containing the parameters used to make the
        time series is returned.

    random_state : int, RandomState instance or None (default = None)
        Determines random number generation for dataset creation. Pass an int
        for reproducible output across multiple function calls.

    Returns
    -------
    X : array, shape = (n_samples, 128)
        Generated time series.

    y : array, shape = (n_samples,)
        Generated class labels. Labels have the following meaning:

        - 0: Cylinder
        - 1: Bell
        - 2: Funnel

    params : dict
        The parameters used to generate the data. The keys are 'a',
        'b', 'eta' and 'epsilon' and the values are the corresponding
        values. Only returned if ``return_params=True``.

    Notes
    -----
    The time series are generated from the following distributions:

    .. math::

        c(t) = (6 + \eta) \cdot 1_{[a, b]}(t) + \epsilon(t)

        b(t) = (6 + \eta) \cdot 1_{[a, b]}(t) \cdot (t - a) / (b - a) +
        \epsilon(t)

        f(t) = (6 + \eta) \cdot 1_{[a, b]}(t) \cdot (b - t) / (b - a) +
        \epsilon(t)

    where :math:`t=1,\ldots,128`, :math:`a` is an integer-valued uniform random
    variable on the inverval :math:`[16, 32]`, :math:`b-a` is an integer-valued
    uniform distribution on the inveral :math:`[32, 96]`, :math:`\eta` and
    :math:`\epsilon(t)` are standard normal variables,
    :math:`{1}_{[a, b]}` is the characteristic function on the interval
    :math:`[a, b]`. :math:`c`, :math:`b`, and :math:`f` stand for "cylinder",
    "bell", and "funnel" respectively.

    References
    ----------
    .. [1] N. Saito, "Local feature extraction and its application
           using a library of bases". Ph.D. thesis, Department of
           Mathematics, Yale University, 1994.

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.datasets import make_cylinder_bell_funnel
    >>> X, y = make_cylinder_bell_funnel()
    >>> X.shape
    (30, 128)
    >>> print(np.bincount(y))
    [10 10 10]

    """
    if not isinstance(n_samples, (int, np.integer)):
        raise TypeError("'n_samples' must be an integer.")
    if not n_samples > 0:
        raise ValueError("'n_samples' must be a positive integer.")
    if weights is None:
        weights = [1.0 / 3] * 3
        weights[-1] = 1.0 - sum(weights[:-1])
    else:
        weights = check_array(weights, ensure_2d=False)
        if len(weights) not in [2, 3]:
            raise ValueError("'weights' must be None or a list with 2 or "
                             "3 elements (got {}).".format(len(weights)))
        if len(weights) == 2:
            if sum(weights) > 1:
                raise ValueError(
                    "'sum(weights)' cannot be larger than 1 if "
                    "len(weights) == 2 (got {})".format(sum(weights)))
            weights = np.append(weights, 1.0 - sum(weights))
    rng = check_random_state(random_state)

    n_samples_per_class = [int(n_samples * weights[k]) for k in range(3)]
    for i in range(n_samples - sum(n_samples_per_class)):
        n_samples_per_class[i] += 1
    n_samples = sum(n_samples_per_class)

    y = ([0] * n_samples_per_class[0] +
         [1] * n_samples_per_class[1] +
         [2] * n_samples_per_class[2])
    y = np.asarray(y)

    a = rng.randint(16, 33)
    b = rng.randint(32, 97) + a
    eta = rng.randn()
    epsilon = rng.randn(n_samples, 128)

    arange = np.tile(np.arange(1, 129), n_samples).reshape(n_samples, 128)
    additional_term = (6 + eta) * np.logical_and(arange >= a, arange <= b)
    i, j = n_samples_per_class[0], sum(n_samples_per_class[0:2])
    additional_term[i:j] *= (np.arange(1, 129) - a) / (b - a)
    additional_term[j:] *= (b - np.arange(1, 129)) / (b - a)

    X = epsilon + additional_term

    if shuffle:
        X, y = utils_shuffle(X, y, random_state=rng)

    if return_params:
        params = {'a': a, 'b': b, 'eta': eta, 'epsilon': epsilon}
        return X, y, params
    else:
        return X, y



"""Code for Piecewise Aggregate Approximation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from numba import njit, prange
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin
from ..utils import segmentation


@njit()
def _paa(X, n_samples, n_timestamps, start, end, n_timestamps_new):
    X_paa = np.empty((n_samples, n_timestamps_new))
    for i in prange(n_samples):
        for j in prange(n_timestamps_new):
            X_paa[i, j] = np.mean(X[i, start[j]:end[j]])
    return X_paa


class PiecewiseAggregateApproximation(BaseEstimator,
                                      UnivariateTransformerMixin):
    """Piecewise Aggregate Approximation.

    Parameters
    ----------
    window_size : int, float or None (default = 1)
        Length of the sliding window. If float, it represents
        a percentage of the size of each time series and must be
        between 0 and 1.

    output_size : int, float or None (default = None)
        Size of the returned time series. If float, it represents
        a percentage of the size of each time series and must be
        between 0. and 1. Ignored if ``window_size`` is not None.
        It can't be None if ``window_size`` is None. If you want to use
        ``output_size`` over ``window_size``, you must set
        ``window_size=None``.

    overlapping : bool (default = True)
        When ``window_size=None``, ``output_size`` is used to derive the window
        size; the window size is fixed if ``overlapping=True`` and may vary
        if ``overlapping=False``. Ignored if ``window_size`` is specified.

    References
    ----------
    .. [1] E. Keogh, K. Chakrabarti, M. Pazzani, and S. Mehrotra,
           "Dimensionality reduction for fast similarity search in large
           time series databases". Knowledge and information Systems,
           3(3), 263-286 (2001).

    Examples
    --------
    >>> from pyts.approximation import PiecewiseAggregateApproximation
    >>> X = [[0, 4, 2, 1, 7, 6, 3, 5],
    ...      [2, 5, 4, 5, 3, 4, 2, 3]]
    >>> transformer = PiecewiseAggregateApproximation(window_size=2)
    >>> transformer.transform(X)
    array([[2. , 1.5, 6.5, 4. ],
           [3.5, 4.5, 3.5, 2.5]])

    """

    def __init__(self, window_size=1, output_size=None, overlapping=True):
        self.window_size = window_size
        self.output_size = output_size
        self.overlapping = overlapping

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        Returns
        -------
        self
            object

        """
        return self

    def transform(self, X):
        """Reduce the dimensionality of each time series.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)

        Returns
        -------
        X_new : array, shape = (n_samples, n_timestamps_new)

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape

        window_size, output_size = self._check_params(n_timestamps)
        if window_size == 1:
            return X
        else:
            start, end, n_timestamps_new = segmentation(
                n_timestamps, window_size, self.overlapping, output_size
            )
            X_paa = _paa(X, n_samples, n_timestamps,
                         start, end, n_timestamps_new)
            return X_paa

    def _check_params(self, n_timestamps):
        if (self.window_size is None and self.output_size is None):
            raise TypeError("'window_size' and 'output_size' cannot be "
                            "both None.")
        if self.window_size is not None:
            if not isinstance(self.window_size,
                              (int, np.integer, float, np.floating)):
                raise TypeError("If specified, 'window_size' must be an "
                                "integer or a float.")
            if isinstance(self.window_size, (int, np.integer)):
                if not 1 <= self.window_size <= n_timestamps:
                    raise ValueError(
                        "If 'window_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "n_timestamps (got {0}).".format(self.window_size)
                    )
                window_size = self.window_size
            else:
                if not 0 < self.window_size <= 1:
                    raise ValueError(
                        "If 'window_size' is a float, it must be greater "
                        "than 0 and lower than or equal to 1 "
                        "(got {0}).".format(self.window_size)
                    )
                window_size = ceil(self.window_size * n_timestamps)
            output_size = None
        else:
            if not isinstance(self.output_size,
                              (int, np.integer, float, np.floating)):
                raise TypeError("If specified, 'output_size' must be an "
                                "integer or a float.")
            if isinstance(self.output_size, (int, np.integer)):
                if not 1 <= self.output_size <= n_timestamps:
                    raise ValueError(
                        "If 'output_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "n_timestamps (got {0}).".format(self.output_size)
                    )
                output_size = self.output_size
            else:
                if not 0 < self.output_size <= 1.:
                    raise ValueError(
                        "If 'output_size' is a float, it must be greater "
                        "than 0 and lower than or equal to 1 "
                        "(got {0}).".format(self.output_size)
                    )
                output_size = ceil(self.output_size * n_timestamps)
            window_size, remainder = divmod(n_timestamps, output_size)
            if remainder != 0:
                window_size += 1
        return window_size, output_size



import numpy as np
import pytest
import re
from pyts.transformation import ShapeletTransform
from pyts.transformation.shapelet_transform import (
    _extract_all_shapelets, _derive_shapelet_distances, _derive_all_distances,
    _remove_similar_shapelets
)

rng = np.random.RandomState(42)
X = rng.randn(4, 8)
y = rng.randint(2, size=4)
X_large = rng.randn(11, 8)
y_large = rng.randint(2, size=11)


@pytest.mark.parametrize(
    'params, res_desired',
    [({'x': np.arange(5), 'window_sizes': (2,),
       'window_steps': (1,), 'n_timestamps': 5},
      {'shapelets': ([[0, 1], [1, 2], [2, 3], [3, 4]],),
       'lengths': (np.full(4, 2),), 'start_idx': [np.arange(4)],
       'end_idx': [np.arange(2, 6)]}),

     ({'x': np.arange(5), 'window_sizes': (2, 3),
       'window_steps': (1, 2), 'n_timestamps': 5},
      {'shapelets': ([[0, 1], [1, 2], [2, 3], [3, 4]], [[0, 1, 2], [2, 3, 4]]),
       'lengths': ([2, 2, 2, 2], [3, 3]), 'start_idx': [(0, 1, 2, 3), (0, 2)],
       'end_idx': [(2, 3, 4, 5), (3, 5)]}),

     ({'x': np.arange(4)[::-1], 'window_sizes': (1, 3),
       'window_steps': (2, 1), 'n_timestamps': 4},
      {'shapelets': ([[3], [1]], [[3, 2, 1], [2, 1, 0]]),
       'lengths': ([1, 1], [3, 3]), 'start_idx': [(0, 2), (0, 1)],
       'end_idx': [(1, 3), (3, 4)]})]
)
def test_extract_all_shapelets(params, res_desired):
    """Test that the actual results are the expected ones."""
    res_actual = _extract_all_shapelets(**params)

    for arr_actual, arr_desired in zip(res_actual[0],
                                       res_desired['shapelets']):
        np.testing.assert_array_equal(arr_actual, arr_desired)

    for arr_actual, arr_desired in zip(res_actual[1], res_desired['lengths']):
        np.testing.assert_array_equal(arr_actual, arr_desired)

    for arr_actual, arr_desired in zip(res_actual[2],
                                       res_desired['start_idx']):
        np.testing.assert_array_equal(arr_actual, arr_desired)

    for arr_actual, arr_desired in zip(res_actual[3], res_desired['end_idx']):
        np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'X, shapelet, arr_desired',
    [(np.ones((5, 3, 2)), np.ones(2), np.zeros(5)),
     (np.ones((5, 3, 2)), np.zeros(2), np.ones(5)),
     (2 * np.ones((5, 3, 2)), np.zeros(2), 4 * np.ones(5)),
     (np.arange(30).reshape(2, 3, 5), np.arange(-2, 3), [4, 289]),
     (np.arange(30).reshape(2, 3, 5), np.zeros(5), [6, 291])]
)
def test_derive_shapelet_distances(X, shapelet, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _derive_shapelet_distances(X, shapelet)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': np.arange(6, dtype='float64').reshape(2, 3), 'window_sizes': (2,),
       'shapelets': (np.array([[1, 2], [2, 1]]),),
       'lengths': (np.array([2, 2]),), 'fit': True},
      np.sqrt([[0, 1], [4, 5]])),

     ({'X': np.arange(8, dtype='float64').reshape(2, 4),
       'window_sizes': (2, 3), 'lengths': np.array([2, 3]), 'fit': False,
       'shapelets': (np.array([1, 2]), np.array([2, 1, 2]))},
      np.sqrt([[0, 1], [9, 12]]))
     ]
)
def test_derive_all_distances(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _derive_all_distances(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'scores': np.arange(10), 'start_idx': np.arange(10),
       'end_idx': np.arange(2, 12)}, [9, 7, 5, 3, 1]),
     ({'scores': np.arange(10), 'start_idx': np.arange(10),
       'end_idx': np.arange(3, 13)}, [9, 6, 3, 0]),
     ({'scores': np.arange(10), 'start_idx': np.arange(10),
       'end_idx': np.arange(4, 14)}, [9, 5, 1]),
     ({'scores': np.arange(10)[::-1], 'start_idx': np.arange(10),
       'end_idx': np.arange(4, 14)}, [0, 4, 8])]
)
def test_remove_similar_shapelets(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _remove_similar_shapelets(**params)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_shapelets': 'yolo'}, TypeError,
      "'n_shapelets' must be 'auto' or an integer."),

     ({'n_shapelets': -2}, ValueError,
      "If 'n_shapelets' is an integer, it must be a positive integer "
      "(got -2)."),

     ({'criterion': 'mse'}, ValueError,
      "'criterion' must be either 'mutual_info' or 'anova' (got mse)."),

     ({'window_sizes': 'mse'}, TypeError,
      "'window_sizes' must be 'auto', a list, a tuple or a numpy.ndarray."),

     ({'window_sizes': [[2, 3, 4]]}, ValueError,
      "'window_sizes' must be one-dimensional."),

     ({'window_sizes': np.array([2, 3]) + 1j}, ValueError,
      "The elements of 'window_sizes' must be integers or floats."),

     ({'window_sizes': [0.5, 2.0]}, ValueError,
      "If the elements of 'window_sizes' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [0.5, -0.5]}, ValueError,
      "If the elements of 'window_sizes' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [3, 40]}, ValueError,
      "If the elements of 'window_sizes' are integers, they all must be "
      "greater than 0 and lower than or equal to n_timestamps."),

     ({'window_sizes': [3, -3]}, ValueError,
      "If the elements of 'window_sizes' are integers, they all must be "
      "greater than 0 and lower than or equal to n_timestamps."),

     ({'window_steps': {0: 0}}, TypeError,
      "'window_steps' must be None or array-like."),

     ({'window_steps': [1, 2]}, ValueError,
      "'window_steps' must be None if window_sizes='auto'."),

     ({'window_sizes': [3, 5], 'window_steps': [[1, 2]]}, ValueError,
      "'window_steps' must be one-dimensional."),

     ({'window_sizes': [3, 5], 'window_steps': [1, 2, 3]}, ValueError,
      "If 'window_steps' is not None, it must have the same size as "
      "'window_sizes'."),

     ({'window_sizes': [3, 5], 'window_steps': [[1, 2]]}, ValueError,
      "'window_steps' must be one-dimensional."),

     ({'window_sizes': [3, 5], 'window_steps': [1, 2 + 1j]}, ValueError,
      "If 'window_steps' is not None, the elements of 'window_steps' must "
      "be integers or floats."),

     ({'window_sizes': [3, 5], 'window_steps': [0.5, 2.0]}, ValueError,
      "If the elements of 'window_steps' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [3, 5], 'window_steps': [0.5, -0.5]}, ValueError,
      "If the elements of 'window_steps' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [3, 5], 'window_steps': [3, 45]}, ValueError,
      "If the elements of 'window_steps' are integers, they all must be "
      "greater than 0 and lower than or equal to n_timestamps."),

     ({'window_sizes': [3, 5], 'window_steps': [3, -3]}, ValueError,
      "If the elements of 'window_steps' are integers, they all must be "
      "greater than 0 and lower than or equal to n_timestamps."),

     ({'verbose': '0'}, ValueError,
      "'verbose' must be a positive integer (got 0)."),

     ({'verbose': -1}, ValueError,
      "'verbose' must be a positive integer (got -1)."),

     ({'n_jobs': '0'}, TypeError,
      "'n_jobs' must be None or an integer.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    shapelet = ShapeletTransform(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        shapelet._check_params(X, y)


@pytest.mark.parametrize(
    'params',
    [{'window_sizes': [3, 5], 'sort': True},
     {'window_sizes': [0.5], 'window_steps': None, 'criterion': 'anova'},
     {'window_sizes': [0.5], 'window_steps': [0.25], 'criterion': 'anova'},
     {}]
)
def test_fit_transform(params):
    """Test that 'fit_transform' and 'fit' then 'transform' yield same res."""
    shapelet_1 = ShapeletTransform(random_state=42, **params)
    shapelet_2 = ShapeletTransform(random_state=42, **params)

    X_fit_transform = shapelet_1.fit_transform(X, y)
    X_fit_then_transform = shapelet_2.fit(X, y).transform(X)

    # Test that the transformation are identical
    np.testing.assert_allclose(X_fit_transform, X_fit_then_transform,
                               atol=1e-5, rtol=0.)

    # Test that the shapelets are identical
    for (shap_1, shap_2) in zip(shapelet_1.shapelets_, shapelet_2.shapelets_):
        np.testing.assert_allclose(shap_1, shap_2, atol=1e-5, rtol=0.)

    # Test that the remaining attributes are identical
    np.testing.assert_allclose(shapelet_1.scores_, shapelet_2.scores_,
                               atol=1e-5, rtol=0.)
    np.testing.assert_array_equal(shapelet_1.indices_, shapelet_2.indices_)
    shapelets_not_none = ((shapelet_1.window_range_ is not None) and
                          (shapelet_2.window_range_ is not None))
    if shapelets_not_none:
        np.testing.assert_array_equal(shapelet_1.window_range_,
                                      shapelet_2.window_range_)


@pytest.mark.parametrize(
    'params, X, y, fewer_shapelets, attr_expected',
    [({'n_shapelets': 5, 'window_sizes': [3]}, X, y, False,
      {'n_shapelets': 5, 'indices_shape': (5, 3), 'scores_size': 5}),

     ({'n_shapelets': 500, 'window_sizes': [3]}, X, y, True,
      {'n_shapelets': 500, 'indices_shape': (500, 3), 'scores_size': 500}),

     ({'n_shapelets': 'auto', 'window_sizes': [3]}, X, y, False,
      {'n_shapelets': 4, 'indices_shape': (4, 3), 'scores_size': 4}),

     ({'n_shapelets': 'auto'}, X, y, False,
      {'n_shapelets': 4, 'indices_shape': (4, 3), 'scores_size': 4}),

     ({'n_shapelets': 'auto'}, X_large, y_large, False,
      {'n_shapelets': 4, 'indices_shape': (4, 3), 'scores_size': 4})]
)
def test_attributes_shape(params, X, y, fewer_shapelets, attr_expected):
    """Test the attributes of a ShapaletTransform instance."""
    shapelet = ShapeletTransform(**params)
    shapelet.fit(X, y)

    # Check 'window_range_'
    window_sizes_auto = (isinstance(shapelet.window_sizes, str) and
                         shapelet.window_sizes == 'auto')
    if window_sizes_auto:
        assert isinstance(shapelet.window_range_, tuple)
        assert (0 < shapelet.window_range_[0]
                <= shapelet.window_range_[1] <= 8)
    else:
        assert shapelet.window_range_ is None

    # Check other attributes
    n_shapelets_actual = len(shapelet.shapelets_)
    if fewer_shapelets:
        assert n_shapelets_actual < attr_expected['n_shapelets']
        assert shapelet.indices_.shape[0] < attr_expected['indices_shape'][0]
        assert shapelet.scores_.size < attr_expected['scores_size']
    else:
        assert n_shapelets_actual == attr_expected['n_shapelets']
        assert shapelet.indices_.shape == attr_expected['indices_shape']
        assert shapelet.scores_.size == attr_expected['scores_size']



"""Code for Shapelet transform algorithm."""

from itertools import chain
from joblib import delayed, Parallel
from numba import njit, prange
from numba.typed import List
import numpy as np
from numpy.lib.stride_tricks import as_strided
from sklearn.base import BaseEstimator
from sklearn.feature_selection import f_classif, mutual_info_classif
from sklearn.utils.validation import (check_array, check_is_fitted,
                                      check_random_state, check_X_y)
from ..base import UnivariateTransformerMixin
from ..utils.utils import _windowed_view


@njit()
def _extract_all_shapelets(x, window_sizes, window_steps, n_timestamps):
    """Extract all the shapelets from a single time series."""
    shapelets = List()  # shapelets
    lengths = List()  # lengths of shapelets
    start_idx = List()  # start index of shapelets (included)
    end_idx = List()  # end index of shapelets (excluded)

    for window_size, window_step in zip(window_sizes, window_steps):
        # Derive the new shape and strides
        overlap = window_size - window_step
        shape_new = ((n_timestamps - overlap) // window_step, window_size // 1)
        strides = x.strides[0]
        strides_new = (window_step * strides, strides)

        # Derive strided view of x
        x_strided = as_strided(x, shape=shape_new, strides=strides_new)
        x_strided = np.copy(x_strided)

        # Add shapelets, lengths, start indices and end indices
        shapelets.append(x_strided)
        lengths.append([x_strided.shape[1]] * x_strided.shape[0])
        start_idx.append(np.arange(0, n_timestamps - window_size + 1,
                                   window_step))
        end_idx.append(np.arange(window_size, n_timestamps + 1, window_step))

    return shapelets, lengths, start_idx, end_idx


@njit()
def _derive_shapelet_distances(X, shapelet):
    """Derive the distance between a shapelet and all the time series."""
    n_samples, n_windows, _ = X.shape
    mean = np.empty((n_samples, n_windows))
    for i in prange(n_samples):
        for j in prange(n_windows):
            mean[i, j] = np.mean((X[i, j] - shapelet) ** 2)
    dist = np.empty(n_samples)
    for i in prange(n_samples):
        dist[i] = np.min(mean[i])
    return dist


@njit()
def _derive_all_squared_distances_fit(
    X, n_samples, n_timestamps, window_sizes, shapelets, lengths
):
    """Derive the squared distances between all shapelets and time series."""
    distances = []  # save the distances in a list

    for i in prange(len(lengths)):
        window_size = lengths[i][0]
        X_window = _windowed_view(X, n_samples, n_timestamps,
                                  window_size, window_step=1)
        for j in prange(shapelets[i].shape[0]):
            dist = _derive_shapelet_distances(X_window, shapelets[i][j])
            distances.append(dist)

    return distances


@njit()
def _derive_all_squared_distances_transform(
    X, n_samples, n_timestamps, window_sizes, shapelets, lengths
):
    """Derive the squared distances between all shapelets and time series."""
    distances = []  # save the distances in a list
    permutation = []  # save the permutation of the indices

    for shapelet, window_size in zip(shapelets, window_sizes):
        X_window = _windowed_view(
            X, n_samples, n_timestamps, window_size, window_step=1
        )
        indices = np.where(lengths == window_size)[0]
        permutation.append(indices)

        for idx in indices:
            dist = _derive_shapelet_distances(X_window, shapelets[idx])
            distances.append(dist)

    return distances, permutation


def _derive_all_distances(X, window_sizes, shapelets, lengths, fit):
    """Derive the distances between all the shapelets and the time series."""
    n_samples, n_timestamps = X.shape
    if fit:
        distances = _derive_all_squared_distances_fit(
            X, n_samples, n_timestamps, window_sizes, shapelets, lengths
        )
        return np.sqrt(np.asarray(distances)).T
    else:
        distances, permutation = _derive_all_squared_distances_transform(
            X, n_samples, n_timestamps, window_sizes, shapelets, lengths
        )
        # Compute the inverse permutation of the indices
        permutation = np.concatenate(permutation)
        inverse_perm = np.arange(permutation.size)[np.argsort(permutation)]

        return np.sqrt(np.asarray(distances))[inverse_perm].T


@njit()
def _remove_similar_shapelets(scores, start_idx, end_idx):
    """Remove self-similar shapelets."""

    kept_idx = []
    remaining_idx = np.full(scores.size, True)

    # Sort the indices by scores
    argsorted_scores = np.argsort(scores)
    sorted_start_idx = start_idx[argsorted_scores]
    sorted_end_idx = end_idx[argsorted_scores]

    # While there are non-similar shapelets remaining
    while np.any(remaining_idx):
        idx = argsorted_scores[remaining_idx][-1]  # Find the best shapelet
        kept_idx.append(idx)

        start, end = start_idx[idx], end_idx[idx]  # Find start and end indices

        # Find non-similar shapelets
        remaining_idx = np.logical_and(
            np.logical_or(sorted_start_idx >= end, sorted_end_idx <= start),
            remaining_idx
        )

    return np.array(kept_idx)


class ShapeletTransform(BaseEstimator, UnivariateTransformerMixin):
    """Shapelet Transform Algorithm.

    The Shapelet Transform algorithm extracts the most discriminative
    shapelets from a data set of time series. A shapelet is defined as
    a subset of consecutive points from a time series. Two criteria are
    made available: mutual information and F-scores.

    Parameters
    ----------
    n_shapelets : int or 'auto' (default = 'auto')
        The number of shapelets to keep. If 'auto', `n_timestamps // 2`
        shapelets are considered, where `n_timestamps` is the number of
        time points in the dataset. Note that there might be a smaller
        number of shapelets if fewer than ``n_shapelets`` shapelets have been
        extracted during the search.

    criterion : 'mutual_info' or 'anova' (default = 'mutual_info')
        Criterion to perform the selection of the shapelets.
        'mutual_info' uses the mutual information, while 'anova' use
        the ANOVA F-value.

    window_sizes : array-like or 'auto '(default = 'auto')
        Size of the sliding windows. If 'auto', the range for the
        window sizes is determined automatically.
        Otherwise, all the elements must be either integers or floats.
        In the latter case, each element represents the percentage
        of the size of each time series and must be between 0 and 1; the size
        of the sliding windows will be computed as
        ``np.ceil(window_sizes * n_timestamps)``.

    window_steps : None or array-like (default = None)
        Step of the sliding windows. If None, each ``window_step`` is equal
        to 1. Otherwise, all the elements must be either integers or floats.
        In the latter case, each element represents the percentage of the size
        of each time series and must be between 0 and 1; the step of the
        sliding windows will be computed as
        ``np.ceil(window_steps * n_timestamps)``.
        Must be None if ``window_sizes='auto'``.

    remove_similar : bool (default = True)
        If True, self-similar shapelets are removed, keeping only
        the non-self-similar shapelets with the highest scores. Two
        shapelets are considered to be self-similar if they are
        taken from the the same time series and have at least one
        overlapping index.

    sort : bool (default = False)
        If True, shapelets are sorted in descending order according to
        their associated scores. If False, the order is undefined.

    verbose : int (default = 0)
        Verbosity level when fitting: if non zero, progress messages are
        printed. Above 50, the output is sent to stdout.
        The frequency of the messages increases with the verbosity level.

    random_state : int, RandomState instance or None (default = None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by ``np.random``. Only used if ``window_sizes='auto'`` in order to
        subsample the dataset to find the best range or if
        ``criterion=='mutual_info'`` to add small noise to the data.

    n_jobs : None or int (default = None)
        The number of jobs to run in parallel for ``fit``.
        If -1, then the number of jobs is set to the number of cores.

    Attributes
    ----------
    shapelets_ : array, shape = (n_shapelets,)
        The array with the selected shapelets.

    indices_ : array, shape = (n_shapelets, 3)
        The indices for the corresponding shapelets in the training set.
        The first column consists of the indices of the samples.
        The second column consists of the starting indices (included)
        of the shapelets. The third column consists of the ending indices
        (excluded) of the shapelets.

    scores_ : array, shape = (n_shapelets,)
        The scores associated to the shapelets. The higher, the more
        discriminant.
        If ``criterion='mutual_info'``, mutual information scores are reported.
        If ``criterion='anova'``, F-scores are reported.

    window_range_ : None or tuple
        Range of the window sizes if ``window_sizes='auto'``. None otherwise.

    References
    ----------
    .. [1] J. Lines, L. M. Davis, J. Hills and A. Bagnall, "A Shapelet
           Transform for Time Series Classification". Data Mining and Knowledge
           Discovery, 289-297 (2012).

    Examples
    --------
    >>> from pyts.transformation import ShapeletTransform
    >>> X = [[0, 2, 3, 4, 3, 2, 1],
    ...      [0, 1, 3, 4, 3, 4, 5],
    ...      [2, 1, 0, 2, 1, 5, 4],
    ...      [1, 2, 2, 1, 0, 3, 5]]
    >>> y = [0, 0, 1, 1]
    >>> st = ShapeletTransform(n_shapelets=2, window_sizes=[3])
    >>> st.fit(X, y)
    ShapeletTransform(...)
    >>> len(st.shapelets_)
    2
    >>> st.indices_.shape
    (2, 3)

    """

    def __init__(self, n_shapelets='auto', criterion='mutual_info',
                 window_sizes='auto', window_steps=None,
                 remove_similar=True, sort=False, verbose=0, random_state=None,
                 n_jobs=None):
        self.n_shapelets = n_shapelets
        self.criterion = criterion
        self.window_sizes = window_sizes
        self.window_steps = window_steps
        self.remove_similar = remove_similar
        self.sort = sort
        self.verbose = verbose
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, X, y):
        """Fit the model according to the given training data.

        It finds the ``n_shapelets`` best shapelets in the training set.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(X, y, dtype='float64')
        (n_shapelets, window_sizes,
         window_steps, n_jobs, rng) = self._check_params(X, y)

        X_new, scores, shapelets, indices = self._fit(
            X, y, n_shapelets, window_sizes, window_steps,
            self.remove_similar, n_jobs, rng
        )

        if self.sort:
            idx_sorted = np.argsort(scores)[::-1]
            self.scores_ = scores[idx_sorted]
            self.shapelets_ = shapelets[idx_sorted]
            self.indices_ = indices[idx_sorted]
        else:
            self.scores_ = scores
            self.shapelets_ = shapelets
            self.indices_ = indices

        return self

    def transform(self, X):
        """Transform the provided data.

        It computes the distances between the selected shapelets
        and the samples.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        X_new : array, shape = (n_samples, n_shapelets)
            Distances between the selected shapelets and the samples.

        """
        check_is_fitted(self, ['shapelets_', 'indices_', 'scores_'])
        X = check_array(X, dtype='float64')
        return self._transform(X)

    def fit_transform(self, X, y):
        """Fit the model than transform the given training data.

        It finds the ``n_shapelets`` best shapelets in the training set
        and computes the distances between them and the training set.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        X_new : array, shape = (n_samples, n_shapelets)
            Distances between the selected shapelets and the samples.

        """
        X, y = check_X_y(X, y, dtype='float64')
        (n_shapelets, window_sizes, window_steps,
         n_jobs, rng) = self._check_params(X, y)

        X_new, scores, shapelets, indices = self._fit(
            X, y, n_shapelets, window_sizes, window_steps,
            self.remove_similar, n_jobs, rng
        )

        if self.sort:
            idx_sorted = np.argsort(scores)[::-1]
            self.scores_ = scores[idx_sorted]
            self.shapelets_ = shapelets[idx_sorted]
            self.indices_ = indices[idx_sorted]
            X_new = X_new[:, idx_sorted]
        else:
            self.scores_ = scores
            self.shapelets_ = shapelets
            self.indices_ = indices
        return X_new

    def _check_params(self, X, y):
        n_timestamps = X.shape[1]

        # Checking for 'n_shapelets'
        n_shapelets_int = isinstance(self.n_shapelets, (int, np.integer))
        if not (self.n_shapelets == 'auto' or n_shapelets_int):
            raise TypeError("'n_shapelets' must be 'auto' or an integer.")
        if (n_shapelets_int and not self.n_shapelets > 0):
            raise ValueError("If 'n_shapelets' is an integer, it must be a "
                             "positive integer (got {})."
                             .format(self.n_shapelets))

        # Checking for 'criterion'
        if self.criterion not in ('mutual_info', 'anova'):
            raise ValueError("'criterion' must be either 'mutual_info' or "
                             "'anova' (got {}).".format(self.criterion))

        # Checking for 'window_sizes'
        window_sizes_auto = (isinstance(self.window_sizes, str) and
                             self.window_sizes == 'auto')
        window_sizes_array = isinstance(self.window_sizes,
                                        (list, tuple, np.ndarray))
        if not (window_sizes_auto or window_sizes_array):
            raise TypeError("'window_sizes' must be 'auto', a list, a tuple "
                            "or a numpy.ndarray.")
        if window_sizes_array:
            window_sizes = np.asarray(self.window_sizes)
            if window_sizes.ndim != 1:
                raise ValueError("'window_sizes' must be one-dimensional.")
            if not issubclass(window_sizes.dtype.type,
                              (np.integer, np.floating)):
                raise ValueError("The elements of 'window_sizes' must be "
                                 "integers or floats.")
            if issubclass(window_sizes.dtype.type, np.floating):
                if not (np.min(window_sizes) > 0 and
                        np.max(window_sizes) <= 1):
                    raise ValueError(
                        "If the elements of 'window_sizes' are floats, they "
                        "all must be greater than 0 and lower than or equal "
                        "to 1."
                    )
                window_sizes = np.ceil(
                    window_sizes * n_timestamps).astype('int64')
            else:
                if not (np.min(window_sizes) > 0 and
                        np.max(window_sizes) <= n_timestamps):
                    raise ValueError(
                        "If the elements of 'window_sizes' are integers, they "
                        "all must be greater than 0 and lower than or equal "
                        "to n_timestamps."
                    )

        # Checking for 'window_steps'
        if not ((self.window_steps is None)
                or isinstance(self.window_steps, (list, tuple, np.ndarray))):
            raise TypeError("'window_steps' must be None or array-like.")
        if self.window_steps is not None:
            if window_sizes_auto:
                raise ValueError(
                    "'window_steps' must be None if window_sizes='auto'."
                )
            window_steps = np.asarray(self.window_steps)
            if window_steps.ndim != 1:
                raise ValueError("'window_steps' must be one-dimensional.")
            if window_steps.size != window_sizes.size:
                raise ValueError("If 'window_steps' is not None, it must have "
                                 "the same size as 'window_sizes'.")
            if not issubclass(window_steps.dtype.type,
                              (np.integer, np.floating)):
                raise ValueError(
                    "If 'window_steps' is not None, the elements of "
                    "'window_steps' must be integers or floats."
                )
            if issubclass(window_steps.dtype.type, np.floating):
                if not ((np.min(window_steps) > 0
                         and np.max(window_steps) <= 1)):
                    raise ValueError(
                        "If the elements of 'window_steps' are floats, they "
                        "all must be greater than 0 and lower than or equal "
                        "to 1."
                    )
                window_steps = np.ceil(
                    window_steps * n_timestamps).astype('int64')
            else:
                if not ((np.min(window_steps) > 0)
                        and (np.max(window_steps) <= n_timestamps)):
                    raise ValueError(
                        "If the elements of 'window_steps' are integers, they "
                        "all must be greater than 0 and lower than or equal "
                        "to n_timestamps."
                    )

        # Checking for 'verbose'
        if not (isinstance(self.verbose, (int, np.integer))
                and self.verbose >= 0):
            raise ValueError("'verbose' must be a positive integer (got {})."
                             .format(self.verbose))

        # Checking for 'n_jobs'
        if not (self.n_jobs is None or isinstance(self.n_jobs, int)):
            raise TypeError("'n_jobs' must be None or an integer.")

        # Derive 'n_jobs' and 'random_state'
        n_jobs = 1 if self.n_jobs is None else self.n_jobs
        rng = check_random_state(self.random_state)

        # Derive parameter values for 'auto' cases
        if self.n_shapelets == 'auto':
            n_shapelets = X.shape[1] // 2
        else:
            n_shapelets = self.n_shapelets

        if isinstance(self.window_sizes, str):
            if self.window_sizes == 'auto':
                window_sizes = self._auto_length_computation(X, y, rng, n_jobs)
                self.window_range_ = (window_sizes[0], window_sizes[-1])
        else:
            self.window_range_ = None

        if self.window_steps is None:
            window_steps = np.ones_like(window_sizes)

        return n_shapelets, window_sizes, window_steps, n_jobs, rng

    def _fit_one_time_series(
        self, x, X, y, n_timestamps, n_shapelets, window_sizes,
        window_steps, remove_similar, i, rng
    ):
        """Fit one time series."""
        # Extract all shapelets
        shapelets, lengths, start_idx, end_idx = _extract_all_shapelets(
            x, window_sizes, window_steps, n_timestamps)

        # Derive distances between shapelets and time series
        X_dist = _derive_all_distances(
            X, window_sizes, shapelets, lengths, fit=True)

        if self.criterion == 'mutual_info':
            scores = mutual_info_classif(X_dist, y, discrete_features=False,
                                         random_state=rng)
        else:
            scores, _ = f_classif(X_dist, y)

        # Flatten the list of 2D arrays into an array of 1D arrays
        shapelets = [list(shapelet) for shapelet in shapelets]
        shapelets = np.asarray(list(chain.from_iterable(shapelets)),
                               dtype='object')

        # Concatenate the list/tuple of 1D arrays into one 1D array
        start_idx = np.concatenate(start_idx)
        end_idx = np.concatenate(end_idx)

        # Remove similar shapelets
        if remove_similar:
            idx = _remove_similar_shapelets(scores.copy(), start_idx, end_idx)
            scores = scores[idx]
            shapelets = shapelets[idx]
            start_idx = start_idx[idx]
            end_idx = end_idx[idx]
            X_dist = X_dist[:, idx]

        # Keep at most 'n_shapelets'
        if scores.size > n_shapelets - 1:
            idx = np.argpartition(
                scores, scores.size - n_shapelets)[-n_shapelets:]
            scores = scores[idx]
            shapelets = shapelets[idx]
            start_idx = start_idx[idx]
            end_idx = end_idx[idx]
            X_dist = X_dist[:, idx]

        time_series_idx = np.full(scores.size, i)
        return X_dist, scores, shapelets, start_idx, end_idx, time_series_idx

    def _fit(self, X, y, n_shapelets, window_sizes, window_steps,
             remove_similar, n_jobs, rng):
        """Fit all the time series"""
        n_samples, n_timestamps = X.shape

        res = Parallel(n_jobs=n_jobs, verbose=self.verbose)(
            delayed(self._fit_one_time_series)(
                X[i], X, y, n_timestamps, n_shapelets, window_sizes,
                window_steps, remove_similar, i, rng
            )
            for i in range(n_samples))

        (X_dist, scores, shapelets,
         start_idx, end_idx, time_series_idx) = zip(*res)

        # Concatenate the results
        X_dist = np.hstack(X_dist)
        scores = np.concatenate(scores)
        shapelets = np.concatenate(shapelets)
        if shapelets.ndim > 1:
            shapelets = shapelets.astype('float64')
        start_idx = np.concatenate(start_idx)
        end_idx = np.concatenate(end_idx)
        time_series_idx = np.concatenate(time_series_idx)

        # Keep at most 'n_shapelets'
        if scores.size > n_shapelets - 1:
            idx = np.argpartition(
                scores, scores.size - n_shapelets)[-n_shapelets:]
            X_dist = X_dist[:, idx]
            scores = scores[idx]
            shapelets = shapelets[idx]
            start_idx = start_idx[idx]
            end_idx = end_idx[idx]
            time_series_idx = time_series_idx[idx]

        # Derive the 'indices' attributes
        indices = np.empty((scores.size, 3), dtype='int64')
        indices[:, 0] = time_series_idx
        indices[:, 1] = start_idx
        indices[:, 2] = end_idx

        return X_dist, scores, shapelets, indices

    def _transform(self, X):
        lengths = self.indices_[:, 2] - self.indices_[:, 1]
        window_sizes = np.unique(lengths)

        # Convert the array of arrays of shapelets to a numba.typed.List object
        shapelets_list = List()
        for x in self.shapelets_:
            shapelets_list.append(x.astype('float64'))

        X_new = _derive_all_distances(
            X, window_sizes, shapelets_list, lengths, fit=False)
        return X_new

    def _auto_length_computation(self, X, y, rng, n_jobs):
        """Derive the window sizes automatically."""
        n_samples, n_timestamps = X.shape
        window_sizes = np.arange(1, n_timestamps + 1)
        window_steps = np.ones_like(window_sizes)

        if n_samples > 10:
            shapelet_lengths = []
            for i in range(10):
                idx = rng.choice(n_samples, size=10, replace=False)
                X_small, y_small = X[idx], y[idx]
                _, _, shapelets, _ = self._fit(
                    X_small, y_small, 10, window_sizes, window_steps,
                    self.remove_similar, n_jobs, rng
                )
                shapelet_lengths.extend(
                    [len(shapelet) for shapelet in shapelets])

            window_range = np.percentile(
                shapelet_lengths, [25, 75], interpolation='lower'
            ).astype('int64')
            window_sizes = np.arange(window_range[0], window_range[1] + 1)

        else:
            _, _, shapelets, _ = self._fit(
                X, y, 100, window_sizes, window_steps,
                self.remove_similar, n_jobs, rng
            )
            shapelet_lengths = [len(shapelet) for shapelet in shapelets]
            window_range = np.percentile(
                shapelet_lengths, [25, 75], interpolation='lower'
            ).astype('int64')
            window_sizes = np.arange(window_range[0], window_range[1] + 1)

        return window_sizes



"""Testing for Time Series Bag-of-Features."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.classification import TSBF
from pyts.classification.tsbf import IntervalFeatureExtractor
from pyts.classification.tsbf import extract_features, histogram
from sklearn.tree import DecisionTreeClassifier


n_samples, n_timestamps = 5, 20
X_arange = np.arange(n_samples * n_timestamps).reshape(n_samples, n_timestamps)
X_ones = np.ones((n_samples, n_timestamps))
X_oob_proba = np.array(
    [[[0.37, 0.95],
      [0.73, 0.60],
      [0.16, 0.16],
      [0.06, 0.87],
      [0.60, 0.71],
      [0.02, 0.97],
      [0.83, 0.21],
      [0.18, 0.18]],

     [[0.30, 0.52],
      [0.43, 0.29],
      [0.61, 0.14],
      [0.29, 0.37],
      [0.46, 0.79],
      [0.20, 0.51],
      [0.59, 0.05],
      [0.61, 0.17]]]
)
y = [0, 0, 1, 1, 2]


def var(n):
    return np.sqrt((n ** 2 - 1) / 12)


@pytest.mark.parametrize(
    'n, res_desired',
    [(3, 0.816496580927726),
     (6, 1.707825127659933),
     (18, 5.188127472091127)]
)
def test_var(n, res_desired):
    np.testing.assert_allclose(var(n), res_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': X_arange,
       'interval_indices': np.array(
           [[[1, 3, 5], [3, 4, 5], [2, 8, 14]],
            [[3, 5, 7], [4, 5, 6], [8, 14, 20]]]
        ),
       'n_samples': 5, 'n_subseries': 3, 'n_intervals': 3},
      [[1.5, 0.5, 1., 3.5, 0.5, 1., 5.5, 0.5, 1., 3.5, var(6), 1., 7.],
       [3., 0., 0., 4., 0., 0., 5., 0., 0., 4., var(3), 3., 6.],
       [4.5, var(6), 1., 10.5, var(6), 1., 16.5, var(6), 1., 10.5, var(18),
        2., 20.],
       [21.5, 0.5, 1., 23.5, 0.5, 1., 25.5, 0.5, 1., 23.5, var(6), 1., 7.],
       [23., 0., 0., 24., 0., 0., 25., 0., 0., 24., var(3), 3., 6.],
       [24.5, var(6), 1., 30.5, var(6), 1., 36.5, var(6), 1., 30.5, var(18),
        2., 20.],
       [41.5, 0.5, 1., 43.5, 0.5, 1., 45.5, 0.5, 1., 43.5, var(6), 1., 7.],
       [43., 0., 0., 44., 0., 0., 45., 0., 0., 44., var(3), 3., 6.],
       [44.5, var(6), 1., 50.5, var(6), 1., 56.5, var(6), 1., 50.5, var(18),
        2., 20.],
       [61.5, 0.5, 1., 63.5, 0.5, 1., 65.5, 0.5, 1., 63.5, var(6), 1., 7.],
       [63., 0., 0., 64., 0., 0., 65., 0., 0., 64., var(3), 3., 6.],
       [64.5, var(6), 1., 70.5, var(6), 1., 76.5, var(6), 1., 70.5, var(18),
        2., 20.],
       [81.5, 0.5, 1., 83.5, 0.5, 1., 85.5, 0.5, 1., 83.5, var(6), 1., 7.],
       [83., 0., 0., 84., 0., 0., 85., 0., 0., 84., var(3), 3., 6.],
       [84.5, var(6), 1., 90.5, var(6), 1., 96.5, var(6), 1., 90.5, var(18),
        2., 20.]]
      ),

     ({'X': X_ones,
       'interval_indices': np.array(
            [[[3, 6, 9, 12], [9, 10, 11, 12]],
             [[6, 9, 12, 15], [10, 11, 12, 13]]]
         ),
      'n_samples': 5, 'n_subseries': 2, 'n_intervals': 4},
      [[1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 3., 15.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 9., 13.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 3., 15.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 9., 13.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 3., 15.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 9., 13.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 3., 15.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 9., 13.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 3., 15.],
       [1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 0., 1., 0., 9., 13.]])]
)
def test_extract_features(params, arr_desired):
    arr_actual = extract_features(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': X_oob_proba, 'bins': 2, 'n_bins': 2,
       'n_samples': 2, 'n_classes': 2},
      [[5, 3, 3, 5, 0.36875, 0.58125], [5, 3, 5, 3, 0.43625, 0.355]]),

     ({'X': X_oob_proba, 'bins': np.array([0., 0.5, 1.]), 'n_bins': 2,
       'n_samples': 2, 'n_classes': 2},
      [[5, 3, 3, 5, 0.36875, 0.58125], [5, 3, 5, 3, 0.43625, 0.355]]),

     ({'X': X_oob_proba, 'bins': 3, 'n_bins': 3,
       'n_samples': 2, 'n_classes': 2},
      [[4, 2, 2, 3, 1, 4, 0.36875, 0.58125],
       [3, 5, 0, 4, 3, 1, 0.43625, 0.355]]),

     ({'X': X_oob_proba, 'bins': np.linspace(0., 1., 4), 'n_bins': 3,
       'n_samples': 2, 'n_classes': 2},
      [[4, 2, 2, 3, 1, 4, 0.36875, 0.58125],
       [3, 5, 0, 4, 3, 1, 0.43625, 0.355]]),

     ({'X': X_oob_proba, 'bins': np.array([0., 0.2, 0.8, 1.]), 'n_bins': 3,
       'n_samples': 2, 'n_classes': 2},
      [[4, 3, 1, 2, 3, 3, 0.36875, 0.58125],
       [0, 8, 0, 3, 5, 0, 0.43625, 0.355]]),

     ({'X': X_oob_proba[:1], 'bins': np.array([0., 0.2, 0.8, 1.]), 'n_bins': 3,
       'n_samples': 1, 'n_classes': 2},
      [[4, 3, 1, 2, 3, 3, 0.36875, 0.58125]]),

     ({'X': X_oob_proba[1:], 'bins': np.array([0., 0.2, 0.8, 1.]), 'n_bins': 3,
       'n_samples': 1, 'n_classes': 2},
      [[0, 8, 0, 3, 5, 0, 0.43625, 0.355]])]
)
def test_histogram(params, arr_desired):
    arr_actual = histogram(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'min_subsequence_size': '4'}, TypeError,
      "'min_subsequence_size' must be an integer or a float."),

     ({'min_subsequence_size': None}, TypeError,
      "'min_subsequence_size' must be an integer or a float."),

     ({'min_subsequence_size': 0}, ValueError,
      "If 'min_subsequence_size' is an integer, it must be greater than or "
      "equal to 1 and lower than or equal to n_timestamps (got 0)."),

     ({'min_subsequence_size': -3}, ValueError,
      "If 'min_subsequence_size' is an integer, it must be greater than or "
      "equal to 1 and lower than or equal to n_timestamps (got -3)."),

     ({'min_subsequence_size': -0.5}, ValueError,
      "If 'min_subsequence_size' is a float, it must be greater "
      "than 0 and lower than or equal to 1 (got -0.5)."),

     ({'min_subsequence_size': -2.}, ValueError,
      "If 'min_subsequence_size' is a float, it must be greater "
      "than 0 and lower than or equal to 1 (got -2.0)."),

     ({'min_interval_size': '4'}, TypeError,
      "'min_interval_size' must be an integer or a float."),

     ({'min_interval_size': []}, TypeError,
      "'min_interval_size' must be an integer or a float."),

     ({'min_interval_size': 0}, ValueError,
      "If 'min_interval_size' is an integer, it must be positive (got 0)."),

     ({'min_interval_size': -6}, ValueError,
      "If 'min_interval_size' is an integer, it must be positive (got -6)."),

     ({'min_interval_size': 0.}, ValueError,
      "If 'min_interval_size' is a float, it must be greater than 0 "
      "(got 0.0)."),

     ({'min_interval_size': -1.}, ValueError,
      "If 'min_interval_size' is a float, it must be greater than 0 "
      "(got -1.0)."),

     ({'min_subsequence_size': 2, 'min_interval_size': 3}, ValueError,
      "'min_interval_size' must be lower than or equal to "
      "'min_subsequence_size' (3 > 2)."),

     ({'min_subsequence_size': 0.1, 'min_interval_size': 3}, ValueError,
      "'min_interval_size' must be lower than or equal to "
      "'min_subsequence_size' (3 > 2)."),

     ({'min_subsequence_size': 0.2, 'min_interval_size': 10}, ValueError,
      "'min_interval_size' must be lower than or equal to "
      "'min_subsequence_size' (10 > 4)."),

     ({'min_subsequence_size': 0.2, 'min_interval_size': 0.5}, ValueError,
      "'min_interval_size' must be lower than or equal to "
      "'min_subsequence_size' (10 > 4)."),

     ({'n_subsequences': []}, TypeError,
      "'n_subsequences' must be 'auto', an integer or a float."),

     ({'n_subsequences': None}, TypeError,
      "'n_subsequences' must be 'auto', an integer or a float."),

     ({'n_subsequences': 0}, ValueError,
      "If 'n_subsequences' is an integer, it must be positive (got 0)."),

     ({'n_subsequences': -6}, ValueError,
      "If 'n_subsequences' is an integer, it must be positive (got -6)."),

     ({'n_subsequences': 0.}, ValueError,
      "If 'n_subsequences' is a float, it must be greater than 0 "
      "(got 0.0)."),

     ({'n_subsequences': -1.}, ValueError,
      "If 'n_subsequences' is a float, it must be greater than 0 "
      "(got -1.0).")]
)
def test_parameter_check_interval_feature_extractor(params, error, err_msg):
    """Test parameter validation."""
    fe = IntervalFeatureExtractor(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        fe.fit(X_arange)


@pytest.mark.parametrize(
    'params, interval_indices_shape, min_subsequence_size',
    [({'min_subsequence_size': 10, 'min_interval_size': 2,
       'n_subsequences': 9}, (9, 6), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 0.1,
       'n_subsequences': 9}, (9, 6), 10),
     ({'min_subsequence_size': 0.5, 'min_interval_size': 2,
       'n_subsequences': 9}, (9, 6), 10),
     ({'min_subsequence_size': 0.5, 'min_interval_size': 0.1,
       'n_subsequences': 9}, (9, 6), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 2,
       'n_subsequences': 4}, (4, 6), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 5,
       'n_subsequences': 6}, (6, 3), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 5,
       'n_subsequences': 0.3}, (6, 3), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 2,
       'n_subsequences': 'auto'}, (5, 6), 10),
     ({'min_subsequence_size': 10, 'min_interval_size': 5,
       'n_subsequences': 'auto'}, (2, 3), 10)]
)
def test_attributes_interval_feature_extractor(
    params, interval_indices_shape, min_subsequence_size
):
    """Test the attributes of a fitted instance of IntervalFeatureExtractor."""
    feature_extractor = IntervalFeatureExtractor(**params)
    feature_extractor.fit(X_arange, y)
    assert feature_extractor.interval_indices_.shape == interval_indices_shape
    assert feature_extractor.min_subsequence_size_ == min_subsequence_size


@pytest.mark.parametrize(
    'X, y, params',
    [(X_arange, y, {}),
     (X_arange, y, {'min_subsequence_size': 6, 'min_interval_size': 2}),
     (X_arange, y, {'min_subsequence_size': 0.8, 'min_interval_size': 2}),
     (X_arange, y, {'min_subsequence_size': 8, 'min_interval_size': 0.2}),
     (X_arange, y, {'min_subsequence_size': 0.8, 'min_interval_size': 0.1}),
     (X_arange, y, {'random_state': 42}),
     (X_arange[:3], y[:3], {}),
     (X_ones, y, {}),
     (X_ones, y, {'min_subsequence_size': 6, 'min_interval_size': 2}),
     (X_ones, y, {'min_subsequence_size': 0.8, 'min_interval_size': 2}),
     (X_ones, y, {'min_subsequence_size': 8, 'min_interval_size': 0.2}),
     (X_ones, y, {'min_subsequence_size': 0.8, 'min_interval_size': 0.1}),
     (X_ones, y, {'random_state': 42}),
     (X_ones[:3], y[:3], {})]
)
def test_methods_interval_feature_extractor(X, y, params):
    """Test the supported methods of IntervalFeatureExtractor."""
    n_samples = X.shape[0]
    feature_extractor = IntervalFeatureExtractor(**params).fit(X, y)
    n_subseries = feature_extractor.interval_indices_.shape[0]
    n_intervals = feature_extractor.interval_indices_.shape[1] - 1
    assert feature_extractor.transform(X).shape == (n_samples * n_subseries,
                                                    3 * n_intervals + 4)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'bins': '4'}, TypeError, "'bins' must be an integer or array-like."),

     ({'bins': None}, TypeError, "'bins' must be an integer or array-like."),

     ({'bins': [0, 1, -1]}, ValueError,
      "If 'bins' is array-like, the bin edges must increase monotonically."),

     ({'bins': (0, 1, -1, 0, 1)}, ValueError,
      "If 'bins' is array-like, the bin edges must increase monotonically."),

     ({'bins': np.array([0, -1, 2])}, ValueError,
      "If 'bins' is array-like, the bin edges must increase monotonically.")]
)
def test_parameter_check_tsbf(params, error, err_msg):
    """Test parameter validation."""
    clf = TSBF(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X_arange, y)


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_estimators': 1}, ValueError,
      "At least one sample was never left out during the bootstrap. "
      "Increase the number of trees (n_estimators).")]
)
def test_no_oob_score(params, error, err_msg):
    clf = TSBF(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X_arange, y)


@pytest.mark.parametrize(
    'X, y, params',
    [(X_arange, y, {}),
     (X_arange, y, {'oob_score': True}),
     (X_arange, y, {'n_estimators': 30, 'random_state': 42, 'bins': 5}),
     (X_arange[:3], y[:3], {}),
     (X_ones, y, {'bins': 3}),
     (X_ones, y, {'n_estimators': 15, 'random_state': 42, 'bins': 5}),
     (X_ones[:3], y[:3], {'bins': np.linspace(0, 1, 13)})]
)
def test_attributes_tsbf(X, y, params):
    """Test the attributes of a fitted instance of TSBF."""
    n_samples = X.shape[0]
    n_classes = np.unique(y).size
    n_estimators = params.get('n_estimators',
                              TSBF().get_params()['n_estimators'])
    bins = params.get('bins', TSBF().get_params()['bins'])
    n_bins = bins if isinstance(bins, (int, np.integer)) else len(bins) - 1
    n_features = (n_bins + 1) * n_classes
    clf = TSBF(**params)

    # Training phase
    clf.fit(X, y)
    assert isinstance(clf.base_estimator_, DecisionTreeClassifier)
    np.testing.assert_array_equal(clf.classes_, np.unique(y))
    assert len(clf.estimators_) == n_estimators
    assert clf.feature_importances_.shape == (n_features,)
    assert clf.interval_indices_.ndim == 2
    assert isinstance(clf.min_subsequence_size_, (int, np.integer))
    assert clf.n_features_in_ == n_features
    if params.get('oob_score', TSBF().get_params()['oob_score']):
        assert clf.oob_decision_function_.shape == (n_samples, n_classes)
        assert isinstance(clf.oob_score_, (float, np.floating))
    else:
        assert clf.oob_decision_function_ is None
        assert clf.oob_score_ is None


@pytest.mark.parametrize(
    'X, y, params',
    [(X_arange, y, {}),
     (X_arange, y, {'n_estimators': 30, 'random_state': 42}),
     (X_arange[:3], y[:3], {}),
     (X_ones, y, {}),
     (X_ones, y, {'bins': 3}),
     (X_ones, y, {'bins': np.linspace(0., 1., 4)}),
     (X_ones, y, {'n_estimators': 15, 'random_state': 42}),
     (X_ones[:3], y[:3], {})]
)
def test_methods_tsbf(X, y, params):
    """Test the supported methods of TSBF."""
    n_samples = X.shape[0]
    n_classes = np.unique(y).size
    n_estimators = params.get('n_estimators',
                              TSBF().get_params()['n_estimators'])
    clf = TSBF(**params).fit(X, y)
    assert clf.apply(X).shape == (n_samples, n_estimators)
    assert clf.decision_path(X)[0].shape[0] == n_samples
    assert clf.decision_path(X)[1].shape == (n_estimators + 1,)
    assert clf.predict(X).shape == (n_samples,)
    assert clf.predict_proba(X).shape == (n_samples, n_classes)
    assert isinstance(clf.score(X, y), (float, np.floating))



"""Code for Lower Bounds of Dynamic Time Warping."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import sqrt
from numba import njit, prange
from sklearn.utils import check_array


def _check_consistent_lengths(X, Y):
    n_timestamps_X, n_timestamps_Y = X.shape[-1], Y.shape[-1]
    if not n_timestamps_X == n_timestamps_Y:
        raise ValueError(
            "Found input variables with inconsistent numbers of "
            "timestamps: [{0}, {1}]".format(n_timestamps_X, n_timestamps_Y)
        )


@njit()
def _lower_bound_yi_x_y(x, x_min, x_max, y, y_min, y_max):
    if x_max >= y_max:
        if x_min < y_min:
            sum1 = np.sum(np.square(x[x > y_max] - y_max))
            sum2 = np.sum(np.square(x[x < y_min] - y_min))
            return sqrt(sum1 + sum2)
        elif x_min > y_max:
            return sqrt(max(np.sum(np.square(x - y_max)),
                            np.sum(np.square(y - x_min))))
        else:
            sum1 = np.sum(np.square(x[x > y_max] - y_max))
            sum2 = np.sum(np.square(y[y < x_min] - x_min))
            return sqrt(sum1 + sum2)
    else:
        if y_min < x_min:
            sum1 = np.sum(np.square(y[y > x_max] - x_max))
            sum2 = np.sum(np.square(y[y < x_min] - x_min))
            return sqrt(sum1 + sum2)
        elif y_min > x_max:
            return sqrt(max(np.sum(np.square(y - x_max)),
                            np.sum(np.square(x - y_min))))
        else:
            sum1 = np.sum(np.square(y[y > x_max] - x_max))
            sum2 = np.sum(np.square(x[x < y_min] - y_min))
            return sqrt(sum1 + sum2)


@njit()
def _lower_bound_yi_X_Y(X, X_min, X_max, Y, Y_min, Y_max):
    n_samples_X, _ = X.shape
    n_samples_Y, _ = Y.shape
    X_yi = np.empty((n_samples_X, n_samples_Y))
    for i in prange(n_samples_X):
        for j in prange(n_samples_Y):
            X_yi[i, j] = _lower_bound_yi_x_y(
                X[i], X_min[i], X_max[i], Y[j], Y_min[j], Y_max[j]
            )
    return X_yi


def lower_bound_yi(X_train, X_test):
    """Compute the "LB_Yi" lower bounds between two datasets.

    Parameters
    ----------
    X_train : array-like, shape = (n_samples_train, n_timestamps)
        Training set.

    X_test: : array-like, shape = (n_samples_test, n_timestamps)
        Test set.

    Returns
    -------
    lower_bounds : array, shape = (n_samples_test, n_samples_train)
        "LB_Yi" lower bounds.

    References
    ----------
    .. [1] B. K. Yi et al, "Efficient Retrieval of Similar Time Sequences
           Under Time Warping". International Conference on Data Engineering,
           201-208 (1998).

    Examples
    --------
    >>> X_train = [[5, 4, 3, 2, 1], [1, 8, 4, 3, 2], [6, 3, 5, 4, 7]]
    >>> X_test = [[2, 1, 8, 4, 5]]
    >>> lower_bound_yi(X_train, X_test)
    array([[3.        , 0.        , 2.44...]])

    """
    X_train = check_array(X_train)
    X_test = check_array(X_test)
    _check_consistent_lengths(X_train, X_test)
    X_train_min = np.min(X_train, axis=1)
    X_train_max = np.max(X_train, axis=1)
    X_test_min = np.min(X_test, axis=1)
    X_test_max = np.max(X_test, axis=1)
    lb_yi = _lower_bound_yi_X_Y(X_test, X_test_min, X_test_max,
                                X_train, X_train_min, X_train_max)
    return lb_yi


def lower_bound_kim(X_train, X_test):
    """Compute the "LB_Kim" lower bounds between two datasets.

    Parameters
    ----------
    X_train : array-like, shape = (n_samples_train, n_timestamps)
        Training set.

    X_test: : array-like, shape = (n_samples_test, n_timestamps)
        Test set.

    Returns
    -------
    lower_bounds : array, shape = (n_samples_test, n_samples_train)
        "LB_Kim" lower bounds.

    References
    ----------
    .. [1] S. W. Kim et al, "An Index-Based Approach for Similarity Search
           Supporting Time Warping in Large Sequence Databases". International
           Conference on Data Engineering, 607-614 (2001).

    Examples
    --------
    >>> X_train = [[2, 1, 8, 4, 5], [1, 2, 3, 4, 5]]
    >>> X_test = [[5, 4, 3, 2, 1], [1, 8, 4, 3, 2], [6, 3, 5, 4, 7]]
    >>> lower_bound_kim(X_train, X_test)
    array([[4, 4],
           [3, 3],
           [4, 5]])

    """
    X_train = check_array(X_train)
    X_test = check_array(X_test)
    _check_consistent_lengths(X_train, X_test)
    first = np.abs(X_test[:, 0, None] - X_train[None, :, 0])
    last = np.abs(X_test[:, -1, None] - X_train[None, :, -1])
    max_ = np.abs(np.max(X_test, axis=1)[:, None]
                  - np.max(X_train, axis=1)[None, :])
    min_ = np.abs(np.min(X_test, axis=1)[:, None]
                  - np.min(X_train, axis=1)[None, :])
    lb_kim = np.max(np.asarray([first, last, max_, min_]), axis=0)
    return lb_kim


@njit()
def _warping_envelope_2d(X, n_samples, n_timestamps, region):
    lower = np.empty((n_samples, n_timestamps))
    upper = np.empty((n_samples, n_timestamps))
    for i in prange(n_samples):
        for j in prange(n_timestamps):
            sub_series = X[i, region[0, j]:region[1, j]]
            lower[i, j] = np.min(sub_series)
            upper[i, j] = np.max(sub_series)
    return lower, upper


@njit()
def _warping_envelope_3d(X, n_samples_X, n_samples_Y, n_timestamps, region):
    lower = np.empty((n_samples_X, n_samples_Y, n_timestamps))
    upper = np.empty((n_samples_X, n_samples_Y, n_timestamps))
    for i in prange(n_samples_X):
        for j in prange(n_samples_Y):
            for k in prange(n_timestamps):
                sub_series = X[i, j, region[0, k]:region[1, k]]
                lower[i, j, k] = np.min(sub_series)
                upper[i, j, k] = np.max(sub_series)
    return lower, upper


def _warping_envelope(X, region):
    """Compute the warping envelope.

    Parameters
    ----------
    X : array
        Input data. It must be two- or three-dimensional.

    region : array, shape = (2, n_timestamps)
        Constraint region. The first row consists of the starting indices
        (included) and the second row consists of the ending indices (excluded)
        of the valid rows for each column.

    Returns
    -------
    lower : array
        The lower envelope.

    upper : array
        The upper envelope.

    """
    X = check_array(X, ensure_2d=False, allow_nd=True)
    region = check_array(region, ensure_min_samples=2)
    n_dims = X.ndim
    if n_dims not in (2, 3):
        raise ValueError("X must be a two- or three-dimensional.")
    if n_dims == 2:
        n_samples, n_timestamps = X.shape
        lower, upper = _warping_envelope_2d(
            X, n_samples, n_timestamps, region
        )
    else:
        n_samples_X, n_samples_Y, n_timestamps = X.shape
        lower, upper = _warping_envelope_3d(
            X, n_samples_X, n_samples_Y, n_timestamps, region
        )
    return lower, upper


@njit()
def _clip_2d(X, X_min, X_max, n_samples_X, n_samples_clip, n_timestamps):
    X_clipped = np.empty((n_samples_X, n_samples_clip, n_timestamps))
    for i in prange(n_samples_X):
        X_clipped[i] = np.minimum(np.maximum(X[i], X_min), X_max)
    return X_clipped


@njit()
def _clip_3d(X, X_min, X_max, n_samples_X, n_samples_clip, n_timestamps):
    X_clipped = np.empty((n_samples_X, n_samples_clip, n_timestamps))
    for i in prange(n_samples_X):
        X_clipped[i] = np.minimum(np.maximum(X[i], X_min[:, i]), X_max[:, i])
    return X_clipped


def _clip(X, lower, upper):
    """Clip an array.

    Parameters
    ----------
    X : array, shape = (n_samples, n_timestamps)
        Array to clip.

    lower : array
        Minimum values in the clipped array. It must be
        two- or three-dimensional.

    upper : array
        Maximum values in the clipped array. It must be
        two- or three-dimensional, and have the same shape
        as ``lower``.

    Returns
    -------
    X_clipped : array
        Clipped array.

    """
    X = check_array(X)
    lower = check_array(lower, ensure_2d=False, allow_nd=True)
    upper = check_array(upper, ensure_2d=False, allow_nd=True)
    n_dims = lower.ndim
    if n_dims not in (2, 3):
        raise ValueError("'lower' must be two- or three-dimensional.")
    if not lower.shape == upper.shape:
        raise ValueError(
            "'lower' and 'upper' must have the same shape "
            "({0} != {1})".format(lower.shape, upper.shape)
        )
    if n_dims == 2:
        n_samples_X, n_timestamps = X.shape
        n_samples_clip, _ = lower.shape
        X_clipped = _clip_2d(
            X, lower, upper, n_samples_X, n_samples_clip, n_timestamps
        )
    else:
        n_samples_X, n_samples_Y, n_timestamps = lower.shape
        X_clipped = _clip_3d(
            X, lower, upper, n_samples_Y, n_samples_X, n_timestamps
        )
    return X_clipped


def lower_bound_keogh(X_train, X_test, region):
    r"""Compute the "LB_Keogh" lower bounds between two datasets.

    Parameters
    ----------
    X_train : array-like, shape = (n_samples_train, n_timestamps)
        Training set. The warping envelopes are computed
        on this set.

    X_test: : array-like, shape = (n_samples_test, n_timestamps)
        Test set.

    region : array, shape = (2, n_timestamps)
        Constraint region. The first row consists of the starting indices
        (included) and the second row consists of the ending indices (excluded)
        of the valid rows for each column.

    Returns
    -------
    lower_bounds : array, shape = (n_samples_test, n_samples_train)
        "LB_Keogh" lower bounds.

    Notes
    -----
    The "LB_Keogh" lower bounds are computed as

    .. math:: LB_Keogh(X, Y) = \Vert X - H(X, Y) \Vert_{2}

    where :math:`X` is the test set (``X_test``), :math:`Y` is the
    training set (``X_train``), and :math:`H(X, Y)` is the projection
    of :math:`X` on :math:`Y`.

    References
    ----------
    .. [1] E. Keogh and C. A. Ratanamahatana, "Exact indexing of dynamic
           time warping". Knowledge and Information Systems, 7(3),
           358-386 (2005).

    Examples
    --------
    >>> X_train = [[0, 1, 2, 3], [1, 2, 3, 4]]
    >>> X_test = [[0, 2.5, 3.5, 6]]
    >>> region = [[0, 0, 1, 2], [2, 3, 4, 4]]
    >>> lower_bound_keogh(X_train, X_test, region)
    array([[3.08...  , 2.23...]])

    """
    X_train = check_array(X_train)
    X_test = check_array(X_test)
    _check_consistent_lengths(X_train, X_test)
    lower, upper = _warping_envelope(X_train, region)
    X_proj = _clip(X_test, lower, upper)
    squared_lb_keogh = np.sum(
        (X_test[:, None, :] - X_proj) ** 2, axis=-1
    )
    return np.sqrt(squared_lb_keogh)


def lower_bound_improved(X_train, X_test, region):
    r"""Compute the "LB_Improved" lower bounds between two datasets.

    Parameters
    ----------
    X_train : array-like, shape = (n_samples_train, n_timestamps)
        Training set.

    X_test: : array-like, shape = (n_samples_test, n_timestamps)
        Test set.

    region : array, shape = (2, n_timestamps)
        Constraint region. The first row consists of the starting indices
        (included) and the second row consists of the ending indices (excluded)
        of the valid rows for each column.

    Returns
    -------
    lower_bounds : array, shape = (n_samples_test, n_samples_train)
        "LB_Improved" lower bounds.

    Notes
    -----
    The "LB_Improved" lower bounds are computed as

    .. math::

        LB_Improved(X, Y) = \sqrt\left\( \Vert X - H(X, Y) \Vert_{2}^2
        + \Vert Y - H(Y, H(X, Y)) \Vert_{2}^2 \right\)

    where :math:`X` is the test set (``X_test``), :math:`Y` is the
    training set (``X_train``), and :math:`H(X, Y)` is the projection
    of :math:`X` on :math:`Y`.

    References
    ----------
    .. [1] D. Lemire, "Faster Retrieval with a Two-Pass Dynamic-Time-Warping
           Lower Bound". Pattern Recognition, 42(9), 2169-2180 (2009).

    Examples
    --------
    >>> X_train = [[0, 1, 2, 3], [1, 2, 3, 4]]
    >>> X_test = [[0, 2.5, 3.5, 3.3]]
    >>> region = [[0, 0, 1, 2], [2, 3, 4, 4]]
    >>> lower_bound_improved(X_train, X_test, region)
    array([[0.76..., 1.11...]])

    """
    X_train = check_array(X_train)
    X_test = check_array(X_test)
    _check_consistent_lengths(X_train, X_test)

    # LB Keogh lower bounds
    lower_train, upper_train = _warping_envelope(X_train, region)
    X_test_proj = _clip(X_test, lower_train, upper_train)
    squared_lb_keogh = np.sum(
        (X_test[:, None, :] - X_test_proj) ** 2, axis=-1
    )

    # LB Improved lower bounds
    lower_test, upper_test = _warping_envelope(X_test_proj, region)
    X_train_proj = _clip(X_train, lower_test, upper_test)
    squared_lb_improved = np.sum(
        (X_train[:, None, :] - X_train_proj) ** 2, axis=-1
    )

    return np.sqrt(squared_lb_keogh + squared_lb_improved.T)



"""The :mod:`pyts.multivariate.utils` module includes utility tools."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.utils import check_array


def check_3d_array(X):
    """Check that the input is a three-dimensional array.

    Parameters
    ----------
    X : array-like
        Input data.

    Returns
    -------
    X_new : array
        Input data as an array.

    """
    X = check_array(X, ensure_2d=False, allow_nd=True)
    if X.ndim != 3:
        raise ValueError("X must be 3-dimensional (got {0}).".format(X.ndim))
    return X



"""Testing for imputers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.preprocessing import InterpolationImputer


X = [[np.nan, 1, 2, 3, np.nan, 5, 6, np.nan]]


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'missing_values': np.inf}, ValueError,
      "'missing_values' cannot be infinity."),

     ({'missing_values': "3"}, ValueError,
      "'missing_values' must be an integer, a float, None or np.nan "
      "(got {0!s})".format("3")),

     ({'strategy': 'whoops'}, ValueError,
      "'strategy' must be an integer or one of 'linear', 'nearest', "
      "'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next' "
      "(got {0})".format('whoops'))]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    imputer = InterpolationImputer(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        imputer.transform(X)


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({'missing_values': None}, [[None, 10, 8, None, 4, 2, None]],
      [[12, 10, 8, 6, 4, 2, 0]]),

     ({'missing_values': np.nan}, [[np.nan, 10, 8, np.nan, 4, 2, np.nan]],
      [[12, 10, 8, 6, 4, 2, 0]]),

     ({'missing_values': 45.}, [[45., 10, 8, 45., 4, 2, 45.]],
      [[12, 10, 8, 6, 4, 2, 0]]),

     ({'missing_values': 78}, [[78, 10, 8, 78, 4, 2, 78]],
      [[12, 10, 8, 6, 4, 2, 0]]),

     ({'missing_values': None, 'strategy': 'quadratic'},
      [[None, 9, 4, None, 0, 1, None]], [[16, 9, 4, 1, 0, 1, 4]]),

     ({'missing_values': None, 'strategy': 'previous'},
      [[5, 9, 4, None, 0, 1, None]], [[5, 9, 4, 4, 0, 1, 1]]),

     ({'missing_values': None, 'strategy': 'next'},
      [[None, 9, 4, None, 0, 1, 8]], [[9, 9, 4, 0, 0, 1, 8]])]
)
def test_actual_results(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    imputer = InterpolationImputer(**params)
    arr_actual = imputer.fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, rtol=0, atol=1e-5)



"""Testing for WEASELMUSE class."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from scipy.sparse import csr_matrix

from pyts.multivariate.transformation import WEASELMUSE

n_samples, n_features, n_timestamps, n_classes = 40, 3, 100, 2
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_features, n_timestamps)
y = rng.randint(n_classes, size=n_samples)


@pytest.mark.parametrize(
    'params, type_desired',
    [({'sparse': True}, csr_matrix), ({'sparse': False}, np.ndarray)]
)
def test_output_dtype(params, type_desired):
    """Check that the output dtype is the expected one."""
    transformer = WEASELMUSE(**params)
    output = transformer.fit_transform(X, y)
    assert isinstance(output, type_desired)


@pytest.mark.parametrize(
    'params', [{'sparse': True}, {'sparse': False}]
)
def test_output_ndim(params):
    """Check that the number of dimensions is always 2."""
    transformer = WEASELMUSE(**params)
    ndim_actual = transformer.fit_transform(X, y).ndim
    assert ndim_actual == 2


def test_n_estimators():
    """Check that the number of estimators is the number of features."""
    transformer = WEASELMUSE().fit(X, y)
    assert len(transformer._estimators) == n_features
    assert len(transformer._estimators_diff) == n_features


@pytest.mark.parametrize(
    'params', [{'sparse': True}, {'sparse': False}]
)
def test_fit_transform(params):
    """Check that fit and transform and fit_transform yield same results."""
    transformer = WEASELMUSE(**params)
    arr_1 = transformer.fit(X, y).transform(X)
    arr_2 = transformer.fit_transform(X, y)
    if transformer.sparse:
        assert (arr_1 != arr_2).nnz == 0
    else:
        np.testing.assert_array_equal(arr_1, arr_2)



"""Testing for Bag-of-Patterns."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from pyts.transformation import BagOfPatterns


X = [[0, 2, 1, 3, 4, 2, 1, 0, 3, 1, 2, 0],
     [2, 0, 1, 3, 2, 4, 1, 2, 0, 1, 3, 2]]


@pytest.mark.parametrize(
    'params, vocab_desired, arr_desired',
    [({'window_size': 4, 'word_size': 4, 'sparse': False},
      {0: 'abdc', 1: 'acbd', 2: 'acdb', 3: 'adbc', 4: 'bacd', 5: 'badb',
       6: 'bdab', 7: 'cabd', 8: 'cbad', 9: 'cbda', 10: 'cdba', 11: 'dbca',
       12: 'dcba'},
      [[0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1, 1],
       [2, 1, 0, 0, 0, 0, 2, 2, 0, 1, 0, 1, 0]]),

     ({'window_size': 4, 'n_bins': 2, 'word_size': 4},
      {0: 'aaba', 1: 'aabb', 2: 'abaa', 3: 'abab', 4: 'abba',
       5: 'baab', 6: 'baba', 7: 'bbaa'},
      [[1, 1, 0, 2, 1, 1, 1, 1], [0, 2, 2, 1, 0, 2, 2, 0]]),

     ({'window_size': 6, 'n_bins': 2, 'word_size': 2},
      {0: 'ab', 1: 'ba'}, [[2, 2], [2, 1]]),

     ({'window_size': 6, 'n_bins': 2, 'word_size': 2,
       'numerosity_reduction': False},
      {0: 'ab', 1: 'ba'}, [[3, 4], [4, 3]])]
)
def test_actual_results(params, vocab_desired, arr_desired):
    """Test that the actual results are the expected ones."""
    bop = BagOfPatterns(**params)
    arr_actual = bop.fit_transform(X)
    assert bop.vocabulary_ == vocab_desired
    if isinstance(arr_actual, np.ndarray):
        np.testing.assert_array_equal(arr_actual, arr_desired)
    else:
        np.testing.assert_array_equal(arr_actual.A, arr_desired)


@pytest.mark.parametrize(
    'params',
    [{'window_size': 4, 'word_size': 4},
     {'window_size': 4, 'word_size': 4, 'sparse': False},
     {'window_size': 4, 'word_size': 4, 'n_bins': 2},
     {'window_size': 4, 'word_size': 4, 'numerosity_reduction': False},
     {'window_size': 4, 'word_size': 4, 'norm_mean': 1, 'norm_std': 1},
     {'window_size': 4, 'word_size': 4, 'overlapping': False},
     {'window_size': 4, 'word_size': 4, 'strategy': 'normal'},
     {'window_size': 4, 'word_size': 4, 'window_step': 2},
     {'window_size': 4, 'word_size': 4, 'alphabet': ['d', 'c', 'b', 'a']}]
)
def test_fit_transform(params):
    """Test that fit_transform and fit then transform yield same results."""
    bop_1, bop_2 = BagOfPatterns(**params), BagOfPatterns(**params)
    arr_1 = bop_1.fit_transform(X)
    arr_2 = bop_2.fit(X).transform(X)
    assert bop_1.vocabulary_ == bop_2.vocabulary_
    if isinstance(arr_1, np.ndarray):
        np.testing.assert_array_equal(arr_1, arr_2)
    else:
        np.testing.assert_array_equal(arr_1.A, arr_2.A)



"""Testing for Gramian Angular Field."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.image.gaf import _gasf, _gadf
from pyts.image import GramianAngularField


X = np.arange(9).reshape(1, 9)
pi_6 = np.cos(np.pi / 6).item()


@pytest.mark.parametrize(
    'X_cos, X_sin, arr_desired',
    [([[-1, 0, 1]], [[-1, 0, 1]], [[[0, 0, 0], [0, 0, 0], [0, 0, 0]]]),

     ([[-1, 0, 1]], [[1, 0, 1]], [[[0, 0, -2], [0, 0, 0], [-2, 0, 0]]]),

     ([[-1, 0, 1], [-1, 0, 1]], [[1, 0, 1], [1, 0, 1]],
      [[[0, 0, -2], [0, 0, 0], [-2, 0, 0]],
       [[0, 0, -2], [0, 0, 0], [-2, 0, 0]]])]
)
def test_actual_results_gasf(X_cos, X_sin, arr_desired):
    """Test that the actual results are the expected ones."""
    X_cos = np.asarray(X_cos)
    X_sin = np.asarray(X_sin)
    arr_actual = _gasf(X_cos, X_sin,
                       n_samples=X_cos.shape[0], image_size=X_cos.shape[1])
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'X_cos, X_sin, arr_desired',
    [([[-1, 0, 1]], [[-1, 0, 1]], [[[0, 0, 0], [0, 0, 0], [0, 0, 0]]]),

     ([[-1, 0, 1]], [[1, 0, 1]], [[[0, 0, 2], [0, 0, 0], [-2, 0, 0]]]),

     ([[-1, 0, 1], [-1, 0, 1]], [[1, 0, 1], [1, 0, 1]],
      [[[0, 0, 2], [0, 0, 0], [-2, 0, 0]],
       [[0, 0, 2], [0, 0, 0], [-2, 0, 0]]])]
)
def test_actual_results_gadf(X_cos, X_sin, arr_desired):
    """Test that the actual results are the expected ones."""
    X_cos = np.asarray(X_cos)
    X_sin = np.asarray(X_sin)
    arr_actual = _gadf(X_cos, X_sin,
                       n_samples=X_cos.shape[0], image_size=X_cos.shape[1])
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'image_size': '4'}, TypeError,
      "'image_size' must be an integer or a float."),

     ({'sample_range': [0, 1]}, TypeError,
      "'sample_range' must be None or a tuple."),

     ({'image_size': 0}, ValueError,
      "If 'image_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'image_size': 2.}, ValueError,
      "If 'image_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(2.)),

     ({'sample_range': (-1, 0, 1)}, ValueError,
      "If 'sample_range' is a tuple, its length must be equal to 2."),

     ({'sample_range': (-2, 2)}, ValueError,
      "If 'sample_range' is a tuple, it must satisfy "
      "-1 <= sample_range[0] < sample_range[1] <= 1."),

     ({'method': 'a'}, ValueError,
      "'method' must be either 'summation', 's', 'difference' or 'd'."),

     ({'sample_range': None}, ValueError,
      "If 'sample_range' is None, all the values of X must be between "
      "-1 and 1.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    gaf = GramianAngularField(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        gaf.transform(X)


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({}, [[-1, 0, 1]], [[[1, 0, -1], [0, -1, 0], [-1, 0, 1]]]),

     ({'sample_range': None}, [[-1, 0, 1]],
      [[[1, 0, -1], [0, -1, 0], [-1, 0, 1]]]),

     ({'image_size': 3}, [[-1, 0, 1]],
      [[[1, 0, -1], [0, -1, 0], [-1, 0, 1]]]),

     ({'image_size': 3}, [np.arange(9)],
      [[[1, 0, -1], [0, -1, 0], [-1, 0, 1]]]),

     ({'image_size': 3, 'overlapping': True}, [np.arange(9)],
      [[[1, 0, -1], [0, -1, 0], [-1, 0, 1]]]),

     ({'method': 'd'}, [[-1, 0, 1]],
      [[[0, 1, 0], [-1, 0, 1], [0, -1, 0]]]),

     ({'image_size': 3, 'method': 'd'}, [np.arange(9)],
      [[[0, 1, 0], [-1, 0, 1], [0, -1, 0]]]),

     ({'sample_range': (0, 1)}, [[-1, 0, 1]],
      [[[-1, -pi_6, 0], [-pi_6, -0.5, 0.5], [0, 0.5, 1]]]),

     ({'sample_range': (0, 1), 'method': 'd'}, [[-1, 0, 1]],
      [[[0, 0.5, 1], [-0.5, 0, pi_6], [-1, -pi_6, 0]]])]
)
def test_actual_results(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = GramianAngularField(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_flatten():
    """Test the 'flatten' parameter."""
    arr_false = GramianAngularField().transform(X).reshape(1, -1)
    arr_true = GramianAngularField(flatten=True).transform(X)
    np.testing.assert_allclose(arr_false, arr_true, atol=1e-5, rtol=0.)



"""WEASEL+MUSE algorithm."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np

from scipy.sparse import csr_matrix, hstack
from sklearn.base import BaseEstimator
from sklearn.base import clone
from sklearn.utils.validation import check_is_fitted
from ...base import MultivariateTransformerMixin
from ...transformation import WEASEL
from ..utils import check_3d_array


class WEASELMUSE(BaseEstimator, MultivariateTransformerMixin):
    r"""WEASEL+MUSE algorithm.

    Parameters
    ----------
    word_size : int (default = 4)
        Size of each word.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and 26.

    window_sizes : array-like (default = [0.1, 0.3, 0.5, 0.7, 0.9])
        Size of the sliding windows. All the elements must be either integers
        or floats. In the latter case, each element represents the percentage
        of the size of each time series and must be between 0 and 1; the size
        of the sliding windows will be computed as
        ``np.ceil(window_sizes * n_timestamps)``.

    window_steps : None or array-like (default = None)
        Step of the sliding windows. If None, each ``window_step`` is equal to
        ``window_size`` so that the windows are non-overlapping. Otherwise, all
        the elements must be either integers or floats. In the latter case,
        each element represents the percentage of the size of each time series
        and must be between 0 and 1; the step of the sliding windows will be
        computed as ``np.ceil(window_steps * n_timestamps)``.

    anova : bool (default = False)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    drop_sum : bool (default = True)
        If True, the first Fourier coefficient (i.e. the sum of the subseries)
        is dropped. Otherwise, it is kept.

    norm_mean : bool (default = True)
        If True, center each subseries before scaling.

    norm_std : bool (default = True)
        If True, scale each subseries to unit variance.

    strategy : str (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    chi2_threshold : int or float (default = 2)
        The threshold used to perform feature selection. Only the words with
        a chi2 statistic above this threshold will be kept.

    sparse : bool (default = True)
        Return a sparse matrix if True, else return an array.

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    Attributes
    ----------
    vocabulary_ : dict
        A mapping of features indices to terms. Each value is a string with
        4 values separated by a whitespace:

            - 'o' or 'd': whether the word is extracted from the original \
                time series ('orig') or from the derivates ('diff')
            - int : feature index
            - int : window size
            - str : word


    References
    ----------
    .. [1] P. Schäfer, and U. Leser, "Multivariate Time Series Classification
           with WEASEL+MUSE". Proceedings of ACM Conference, (2017).

    Examples
    --------
    >>> from pyts.datasets import load_basic_motions
    >>> from pyts.multivariate.transformation import WEASELMUSE
    >>> X_train, X_test, y_train, y_test = load_basic_motions(return_X_y=True)
    >>> transformer = WEASELMUSE()
    >>> X_new = transformer.fit_transform(X_train, y_train)
    >>> X_new.shape
    (40, 9086)

    """

    def __init__(self, word_size=4, n_bins=4,
                 window_sizes=[0.1, 0.3, 0.5, 0.7, 0.9], window_steps=None,
                 anova=False, drop_sum=True, norm_mean=True, norm_std=True,
                 strategy='quantile', chi2_threshold=2, sparse=True,
                 alphabet=None):
        self.word_size = word_size
        self.n_bins = n_bins
        self.window_sizes = window_sizes
        self.window_steps = window_steps
        self.anova = anova
        self.drop_sum = drop_sum
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.strategy = strategy
        self.chi2_threshold = chi2_threshold
        self.sparse = sparse
        self.alphabet = alphabet

    def fit(self, X, y):
        """Fit the model according to the given training data.

        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : array-like, shape = (n_samples,)
            Class labels.

        Returns
        -------
        self : object

        """
        X = check_3d_array(X)
        _, n_features, n_timestamps = X.shape

        X_diff = np.abs(np.diff(X))

        estimator = WEASEL(
            word_size=self.word_size, n_bins=self.n_bins,
            window_sizes=self.window_sizes, window_steps=self.window_steps,
            anova=self.anova, drop_sum=self.drop_sum, norm_mean=self.norm_mean,
            norm_std=self.norm_std, strategy=self.strategy,
            chi2_threshold=self.chi2_threshold, sparse=self.sparse,
            alphabet=self.alphabet
        )
        self._estimators = [clone(estimator) for _ in range(n_features)]
        self._estimators_diff = [clone(estimator) for _ in range(n_features)]

        self.vocabulary_ = {}

        for i, transformer in enumerate(self._estimators):
            transformer.fit(X[:, i, :], y)
            self._update_vocabulary(str(i + 1), transformer, original=True)

        for i, transformer in enumerate(self._estimators_diff):
            transformer.fit(X_diff[:, i, :], y)
            self._update_vocabulary(str(i + 1), transformer, original=False)

        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        Returns
        -------
        X_new : sparse matrix, shape = (n_samples, n_features_new)
            Document-term matrix with relevant learned features only.

        """
        check_is_fitted(self, 'vocabulary_')
        X = check_3d_array(X)
        n_samples, _, _ = X.shape
        X_diff = np.abs(np.diff(X))

        X_new = []
        for i, transformer in enumerate(self._estimators):
            X_new.append(transformer.transform(X[:, i, :]))
        for i, transformer in enumerate(self._estimators_diff):
            X_new.append(transformer.transform(X_diff[:, i, :]))

        if self.sparse:
            return csr_matrix(hstack(X_new))
        return np.hstack(X_new)

    def fit_transform(self, X, y):
        """Fit the data then transform it.

        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : array-like, shape = (n_samples,)
            Class labels.

        Returns
        -------
        X_new : array, shape = (n_samples, n_features_new)
            Document-term matrix with relevant features only.

        """
        X = check_3d_array(X)
        n_samples, n_features, n_timestamps = X.shape

        X_diff = np.abs(np.diff(X))

        estimator = WEASEL(
            word_size=self.word_size, n_bins=self.n_bins,
            window_sizes=self.window_sizes, window_steps=self.window_steps,
            anova=self.anova, drop_sum=self.drop_sum, norm_mean=self.norm_mean,
            norm_std=self.norm_std, strategy=self.strategy,
            chi2_threshold=self.chi2_threshold, sparse=self.sparse,
            alphabet=self.alphabet
        )
        self._estimators = [clone(estimator) for _ in range(n_features)]
        self._estimators_diff = [clone(estimator) for _ in range(n_features)]

        self.vocabulary_ = {}

        X_new = []
        for i, transformer in enumerate(self._estimators):
            X_new.append(transformer.fit_transform(X[:, i, :], y))
            self._update_vocabulary(str(i + 1), transformer, original=True)
        for i, transformer in enumerate(self._estimators_diff):
            X_new.append(transformer.fit_transform(X_diff[:, i, :], y))
            self._update_vocabulary(str(i + 1), transformer, original=False)

        if self.sparse:
            return csr_matrix(hstack(X_new))
        return np.hstack(X_new)

    def _update_vocabulary(self, feature_idx, estimator, original):
        """Update the vocabulary."""
        old_length = len(self.vocabulary_)
        if original:
            for (key, value) in estimator.vocabulary_.items():
                self.vocabulary_[old_length + key] = (
                    "o " + feature_idx + " " + value
                )
        else:
            for (key, value) in estimator.vocabulary_.items():
                self.vocabulary_[old_length + key] = (
                    "d " + feature_idx + " " + value
                )



"""Code for Time Series Bag-of-Features."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from math import ceil
from numba import njit
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier
from sklearn.utils.validation import (
    check_array, check_is_fitted, check_random_state, check_X_y
)
from ..base import UnivariateClassifierMixin, UnivariateTransformerMixin


@njit
def extract_features(X, interval_indices, n_samples, n_subseries, n_intervals):
    # There are 3 features for each interval (mean, std, slope) plus
    # the mean, std over the whole subsequence and indices of the subsequence
    n_features = n_intervals * 3 + 4
    X_features = np.zeros((n_samples * n_subseries, n_features))

    for j in range(n_subseries):
        for k in range(n_intervals):

            # Get the start and end indices of the interval
            start, end = interval_indices[:, j, k]
            arange = np.arange((start - end + 1) / 2, (end + 1 - start) / 2)
            if end - start == 1:
                var_arange = 1.
            else:
                var_arange = np.sum(arange ** 2)

            for i in range(n_samples):
                i_j = i * n_subseries + j
                mean = np.mean(X[i, start:end])

                # Compute the three statistics of the interval
                X_features[i_j, 3 * k] = mean
                X_features[i_j, 3 * k + 1] = np.std(X[i, start:end])
                X_features[i_j, 3 * k + 2] = (
                    np.sum((X[i, start:end] - mean) * arange) / var_arange
                )

        start, end = interval_indices[0, j, 0], interval_indices[1, j, -1]
        for i in range(n_samples):
            i_j = i * n_subseries + j

            # Compute the four statistics of the subseries
            X_features[i_j, -4] = np.mean(X[i, start:end])
            X_features[i_j, -3] = np.std(X[i, start:end])
            X_features[i_j, -2] = start
            X_features[i_j, -1] = end

    return X_features


@njit
def histogram(X, bins, n_bins, n_samples, n_classes):
    X_new = np.empty((n_samples, (n_bins + 1) * n_classes))
    for i in range(n_samples):
        for j in range(n_classes):
            X_new[i, j * n_bins: (j + 1) * n_bins] = np.histogram(
                X[i, :, j], bins, (0., 1.))[0]
            X_new[i, - n_classes + j] = np.mean(X[i, :, j])
    return X_new


class IntervalFeatureExtractor(BaseEstimator, UnivariateTransformerMixin):
    """Feature extractor over the intervals of a subsequence.

    This transformer extracts 3 features from each interval of each
    subsequence: the mean, the standard deviation and the slope.

    Parameters
    ----------
    min_subsequence_size : int or float (default = 0.5)
        The minimum length of the subsequences. If float, it represents a
        percentage of the size of each time series. Note that the actual
        minimum length of the subsequences may be slighty lower due to the use
        of the floor function.

    min_interval_size : int or float (default = 0.1)
        The minimum length of the intervals. If float, it represents a
        percentage of the size of each time series. Must be lower than
        `min_subsequence_size`.

    n_subsequences : 'auto', int or float (default = 'auto')
        The number of considered subsequences. If 'auto', it is automatically
        calculated given the size of the time series and the values of the
        parameters. If float, it represents a percentage of the size of each
        time series.

    bins : int or array-like (default = 10)
        If bins is an int, it defines the number of equal-width bins in range
        [0, 1]. If bins is array-like, it defines a monotonically increasing
        array of bin edges, including the rightmost edge, allowing for
        non-uniform bin widths.

    random_state : None, int or RandomState instance (default = None)
        The seed of the pseudo random number generator to use when shuffling
        the data. If int, random_state is the seed used by the random number
        generator. If RandomState instance, random_state is the random number
        generator. If None, the random number generator is the RandomState
        instance used by `np.random`.

    Attributes
    ----------
    interval_indices_ : array, shape = (n_subsequences, n_intervals + 1)
        The indices for the intervals of each subsequence.

    min_subsequence_size_ : int
        The actual minimum length of the subsequences.

    """
    def __init__(self, min_subsequence_size=0.5, min_interval_size=0.1,
                 n_subsequences='auto', random_state=None):
        self.min_subsequence_size = min_subsequence_size
        self.min_interval_size = min_interval_size
        self.n_subsequences = n_subsequences
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        It generates the indices of the windows from which the features will be
        extracted.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y
            Ignored

        Returns
        -------
        self : object

        """
        X = check_array(X, dtype='float64')
        n_timestamps = X.shape[1]

        (min_subsequence_size, min_interval_size, n_subsequences,
         rng) = self._check_params(X, n_timestamps)

        # Define the parameters of the subseries and intervals
        n_intervals = int(min_subsequence_size / min_interval_size)
        min_subseries_size = min_interval_size * n_intervals
        if n_subsequences == 'auto':
            n_subseries = max(
                1, n_timestamps // min_interval_size - n_intervals)
        else:
            n_subseries = n_subsequences

        # Generate the indices of the subseries
        subseries_start = rng.randint(
            0, n_timestamps - min_subseries_size + 1, size=n_subseries
        )
        subseries_end = rng.randint(
            subseries_start + min_subseries_size, n_timestamps + 1,
            size=n_subseries
        )

        # Generate the indices of the intervals
        interval_indices = np.linspace(
            subseries_start, subseries_end, n_intervals + 1, axis=1
        ).astype('int64')

        # Save the relevant parameters as attributes
        self.interval_indices_ = interval_indices
        self.min_subsequence_size_ = min_subsequence_size

        return self

    def transform(self, X):
        """Transform the provided data.

        It extracts the three features from all the selected windows
        for all the samples.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        X_new : array, shape = (n_samples * n_subseries, 3 * n_intervals + 4)
            Extracted features.

        """
        X = check_array(X, dtype='float64')
        n_samples = X.shape[0]

        # Extract the features from each interval
        interval_indices = np.array(
            [self.interval_indices_[:, :-1], self.interval_indices_[:, 1:]]
        )

        n_subseries, n_intervals = interval_indices.shape[1:]
        X_features = extract_features(X, interval_indices, n_samples,
                                      n_subseries, n_intervals)
        return X_features

    def _check_params(self, X, n_timestamps):

        if not isinstance(self.min_subsequence_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'min_subsequence_size' must be an integer or a "
                            "float.")
        if isinstance(self.min_subsequence_size, (int, np.integer)):
            if not 1 <= self.min_subsequence_size <= n_timestamps:
                raise ValueError(
                    "If 'min_subsequence_size' is an integer, it must be "
                    "greater than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.min_subsequence_size)
                )
            min_subsequence_size = self.min_subsequence_size
        else:
            if not (0 < self.min_subsequence_size <= 1.):
                raise ValueError(
                    "If 'min_subsequence_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.min_subsequence_size)
                )
            min_subsequence_size = ceil(
                self.min_subsequence_size * n_timestamps)

        if not isinstance(self.min_interval_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'min_interval_size' must be an integer or a "
                            "float.")
        if isinstance(self.min_interval_size, (int, np.integer)):
            if self.min_interval_size < 1:
                raise ValueError(
                    "If 'min_interval_size' is an integer, it must be "
                    "positive (got {0}).".format(self.min_interval_size)
                )
            min_interval_size = self.min_interval_size
        else:
            if self.min_interval_size <= 0:
                raise ValueError(
                    "If 'min_interval_size' is a float, it must be greater "
                    "than 0 (got {0}).".format(self.min_interval_size)
                )
            min_interval_size = ceil(self.min_interval_size * n_timestamps)

        if min_interval_size > min_subsequence_size:
            raise ValueError("'min_interval_size' must be lower than or equal "
                             "to 'min_subsequence_size' ({} > {})."
                             .format(min_interval_size, min_subsequence_size))

        n_subsequences_auto = (isinstance(self.n_subsequences, str) and
                               self.n_subsequences == 'auto')
        if not (n_subsequences_auto or isinstance(
            self.n_subsequences, (int, np.integer, float, np.floating))
        ):
            raise TypeError("'n_subsequences' must be 'auto', an integer or a "
                            "float.")
        if n_subsequences_auto:
            n_subsequences = 'auto'
        elif isinstance(self.n_subsequences, (int, np.integer)):
            if self.n_subsequences < 1:
                raise ValueError(
                    "If 'n_subsequences' is an integer, it must be positive "
                    "(got {0}).".format(self.n_subsequences)
                )
            n_subsequences = self.n_subsequences
        else:
            if self.n_subsequences <= 0:
                raise ValueError(
                    "If 'n_subsequences' is a float, it must be greater "
                    "than 0 (got {0}).".format(self.n_subsequences)
                )
            n_subsequences = ceil(self.n_subsequences * n_timestamps)

        rng = check_random_state(self.random_state)

        return (min_subsequence_size, min_interval_size,
                n_subsequences, rng)


class TSBF(BaseEstimator, UnivariateClassifierMixin):
    """Time Series Bag-of-Features algorithm.

    Parameters
    ----------
    n_estimators : int (default = 500)
        The number of trees in the forest.

    min_subsequence_size : int or float (default = 0.5)
        The minimum length of the subsequences. If float, it represents a
        percentage of the size of each time series. Note that the actual
        minimum length of the subsequences may be slighty lower due to the use
        of the floor function.

    min_interval_size : int or float (default = 0.1)
        The minimum length of the intervals. If float, it represents a
        percentage of the size of each time series. Must be lower than
        `min_subsequence_size`.

    n_subsequences : 'auto', int or float (default = 'auto')
        The number of considered subsequences. If 'auto', it is automatically
        calculated given the size of the time series and the values of the
        parameters. If float, it represents a percentage of the size of each
        time series.

    bins : int or array-like (default = 10)
        If bins is an int, it defines the number of equal-width bins in range
        [0, 1]. If bins is array-like, it defines a monotonically increasing
        array of bin edges, including the rightmost edge, allowing for
        non-uniform bin widths.

    criterion : str (default = "entropy")
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "entropy" for the information gain.
        Note: this parameter is tree-specific.

    max_depth : integer or None (default = None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        ``min_samples_split`` samples.

    min_samples_split : int or float (default = 2)
        The minimum number of samples required to split an internal node:

        - If int, then consider ``min_samples_split`` as the minimum number.
        - If float, then ``min_samples_split`` is a fraction and
          ``ceil(min_samples_split * n_samples)`` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float (default = 1)
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model.

        - If int, then consider ``min_samples_leaf`` as the minimum number.
        - If float, then ``min_samples_leaf`` is a fraction and
          ``ceil(min_samples_leaf * n_samples)`` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float (default = 0.)
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node.

    max_features : int, float, str or None (default = "auto")
        The number of features to consider when looking for the best split:

        - If int, then consider ``max_features`` features at each split.
        - If float, then ``max_features`` is a fraction and
          ``int(max_features * n_features)`` features are considered at each
          split.
        - If "auto", then ``max_features=sqrt(n_features)``.
        - If "sqrt", then ``max_features=sqrt(n_features)`` (same as "auto").
        - If "log2", then ``max_features=log2(n_features)``.
        - If None, then ``max_features=n_features``.

    max_leaf_nodes : int or None (default = None)
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float (default = 0.)
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.
        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

    bootstrap : bool (default = True)
        Whether bootstrap samples are used when building trees. If False, the
        whole datset is used to build each tree.

    oob_score : bool (default = False)
        Whether to use out-of-bag samples to estimate
        the generalization accuracy.

    n_jobs : int or None, optional (default = None)
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a ``joblib.parallel_backend``
        context. ``-1`` means using all processors.

    random_state : int, RandomState instance or None (default = None)
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).

    verbose : int (default = 0)
        Controls the verbosity when fitting and predicting.

    class_weight : dict, "balanced", "balanced_subsample" or None (default = None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``
        The "balanced_subsample" mode is the same as "balanced" except that
        weights are computed based on the bootstrap sample for every tree
        grown.

    ccp_alpha : float (default = 0.)
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed.
        It must be non-negative.

    max_samples : int, float or None (default = None)
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator:

        - If None (default), then draw ``X.shape[0]`` samples.
        - If int, then draw ``max_samples`` samples.
        - If float, then draw ``max_samples * X.shape[0]`` samples. Thus,
          ``max_samples`` should be in the interval `(0, 1)`.

    Attributes
    ----------
    base_estimator_ : DecisionTreeClassifier
        The child estimator template used to create the collection of fitted
        sub-estimators.

    classes_ : array, shape = (n_classes,)
        The classes labels.

    estimators_ : list of DecisionTreeClassifier
        The collection of fitted sub-estimators.

    feature_importances_ : array, shape = (n_features,)
        The feature importances (the higher, the more important the feature).

    interval_indices_ : array, shape = (n_subsequences, n_intervals + 1)
        The indices for the intervals of each subsequence.

    min_subsequence_size_ : int
        The actual minimum length of the subsequences.

    n_features_in_ : int
        The number of features when ``fit`` is performed.

    oob_decision_function_ : None or array, shape = (n_samples, n_classes)
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN. This attribute is not None
        only when ``oob_score`` is True.

    oob_score_ : None or float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute is not None only when ``oob_score`` is True.

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.classification import TSBF
    >>> X_train, X_test, y_train, y_test = load_gunpoint(return_X_y=True)
    >>> clf = TSBF(random_state=43)
    >>> clf.fit(X_train, y_train)
    TSBF(...)
    >>> clf.score(X_test, y_test)
    0.97333...

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    The features are always randomly permuted at each split. Therefore,
    the best found split may vary, even with the same training data,
    ``max_features=n_features`` and ``bootstrap=False``, if the improvement
    of the criterion is identical for several splits enumerated during the
    search of the best split. To obtain a deterministic behaviour during
    fitting, ``random_state`` has to be fixed.

    References
    ----------
    .. [1] M.G. Baydogan, G. Runger and E. Tuv, "A Bag-of-Features Framework
       to Classify Time Series". IEEE Transactions on Pattern Analysis
       and Machine Intelligence, 35(11), 2796-2802 (2013).

    .. [2] Leo Breiman, "Random Forests", Machine Learning, 45(1), 5-32, 2001.

    """  # noqa: E501
    def __init__(self,
                 n_estimators=500,
                 min_subsequence_size=0.5,
                 min_interval_size=0.1,
                 n_subsequences='auto',
                 bins=10,
                 criterion="entropy",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 class_weight=None,
                 ccp_alpha=0.0,
                 max_samples=None):
        self.n_estimators = n_estimators
        self.min_subsequence_size = min_subsequence_size
        self.min_interval_size = min_interval_size
        self.n_subsequences = n_subsequences
        self.bins = bins
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples

    def apply(self, X):
        """Apply trees in the forest to X, return leaf indices.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        X_leaves : array_like, shape = (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.

        """
        X_binned = self._transform(X)
        return self._clf.apply(X_binned)

    def decision_path(self, X):
        """Return the decision path in the forest.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        indicator : sparse csr array, shape = (n_samples, n_nodes)
            Return a node indicator matrix where non zero elements
            indicates that the samples goes through the nodes.

        n_nodes_ptr : array, shape = (n_estimators + 1,)
            The columns from indicator[n_nodes_ptr[i]:n_nodes_ptr[i+1]]
            gives the indicator value for the i-th estimator.

        """
        X_binned = self._transform(X)
        return self._clf.decision_path(X_binned)

    def fit(self, X, y):
        """Fit the model according to the given training data.

        It build a forest of trees from the training set.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y : array-like, shape = (n_samples,)
            Class labels for each sample.

        Returns
        -------
        self : object

        """
        # Extract some information from the dataset
        X, y = check_X_y(X, y)
        n_samples, n_timestamps = X.shape
        n_classes = np.unique(y).size

        bins = self._check_params()
        self._bins = bins

        # Create the instances of the feature extractor anf classifiers
        feature_extractor = IntervalFeatureExtractor(
            min_subsequence_size=self.min_subsequence_size,
            min_interval_size=self.min_interval_size,
            n_subsequences=self.n_subsequences,
            random_state=self.random_state
        )
        rfc = RandomForestClassifier(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            class_weight=self.class_weight,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples,
            warm_start=False,
            oob_score=True
        )
        clf = RandomForestClassifier(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            class_weight=self.class_weight,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples,
            warm_start=False,
        )

        # Fit the random forest and get out-of-bag probabilities
        X_features = feature_extractor.fit_transform(X)
        n_subsequences = feature_extractor.interval_indices_.shape[0]
        y_features = np.repeat(y, n_subsequences)
        rfc.fit(X_features, y_features)
        X_oob_proba = rfc.oob_decision_function_.reshape(
            n_samples, n_subsequences, n_classes)

        # Check for subsequences without OOB scores
        no_oob_scores = (
            (np.isnan(X_oob_proba).any() or
             np.all(X_oob_proba == 0., axis=2).any())
        )
        if no_oob_scores:
            raise ValueError(
                "At least one sample was never left out during the bootstrap. "
                "Increase the number of trees (n_estimators)."
            )

        # Bin the probabilities
        if isinstance(bins, (int, np.integer)):
            n_bins = bins
        else:
            n_bins = len(bins) - 1
        X_binned = histogram(X_oob_proba, bins, n_bins, n_samples, n_classes)

        # Fit a random forest on the binned probabilities
        clf.fit(X_binned, y)

        # Get attributes
        self.base_estimator_ = clf.base_estimator_
        self.classes_ = clf.classes_
        self.estimators_ = clf.estimators_
        self.feature_importances_ = clf.feature_importances_
        self.interval_indices_ = feature_extractor.interval_indices_
        self.min_subsequence_size_ = feature_extractor.min_subsequence_size_
        self.n_features_in_ = (
            clf.n_features_in_
            if hasattr(clf, 'n_features_in_')
            else clf.n_features_
        )
        self.oob_decision_function_ = getattr(
            clf, 'oob_decision_function_', None)
        self.oob_score_ = getattr(clf, 'oob_score_', None)

        self._feature_extractor = feature_extractor
        self._rfc = rfc
        self._clf = clf

        return self

    def predict(self, X):
        """Predict class for X.

        The predicted class of an input time series is a vote by the trees
        in the forest, weighted by their probability estimates.
        That is, the predicted class is the one with highest mean
        probability estimate across the trees.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        y : array, shape = (n_samples,)
            The predicted classes.

        """
        X_binned = self._transform(X)
        return self._clf.predict(X_binned)

    def predict_proba(self, X):
        """Predict class probabilities for X.

        The predicted class probabilities of an input time series are computed
        as the mean predicted class probabilities of the trees in the forest.
        The class probability of a single tree is the fraction of samples
        of the same class in a leaf.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        Returns
        -------
        p : array, shape = (n_samples, n_classes)
            The class probabilities of the input time series.
            The order of the classes corresponds to that in the
            attribute `classes_`.

        """
        X_binned = self._transform(X)
        return self._clf.predict_proba(X_binned)

    def _check_params(self):
        if not isinstance(self.bins,
                          (int, np.integer, list, tuple, np.ndarray)):
            raise TypeError("'bins' must be an integer or array-like.")
        if isinstance(self.bins, (int, np.integer)):
            bins = self.bins
        else:
            bins = np.asarray(self.bins)
            if (np.diff(bins) < 0).any():
                raise ValueError("If 'bins' is array-like, the bin edges "
                                 "must increase monotonically.")
        return bins

    def _transform(self, X):
        check_is_fitted(self)
        X = check_array(X, dtype='float64')
        n_samples = X.shape[0]
        n_subsequences = self.interval_indices_.shape[0]
        n_classes = len(self.classes_)

        # Predict probabilities using the inner random forest
        X_features = self._feature_extractor.transform(X)
        X_proba = self._rfc.predict_proba(X_features).reshape(
            n_samples, n_subsequences, n_classes)

        # Bin the probabilities
        if isinstance(self._bins, (int, np.integer)):
            n_bins = self._bins
        else:
            n_bins = len(self._bins) - 1
        X_binned = histogram(X_proba, self._bins, n_bins, n_samples, n_classes)

        return X_binned



"""Code for Singular Spectrum Analysis."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from numba import njit, prange
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin
from ..utils.utils import _windowed_view


@njit
def _outer_dot(v, X, n_samples, window_size, n_windows):
    X_new = np.empty((n_samples, window_size, window_size, n_windows))
    for i in prange(n_samples):
        for j in prange(window_size):
            X_new[i, j] = np.dot(np.outer(v[i, :, j], v[i, :, j]), X[i])
    return X_new


@njit
def _diagonal_averaging(X, n_samples, n_timestamps, window_size,
                        n_windows, grouping_size, gap):
    X_new = np.empty((n_samples, grouping_size, n_timestamps))
    first_row = [(0, col) for col in range(n_windows)]
    last_col = [(row, n_windows - 1) for row in range(1, window_size)]
    indices = first_row + last_col
    for i in prange(n_samples):
        for group in prange(grouping_size):
            for (j, k) in indices:
                X_new[i, group, j + k] = np.diag(
                    X[i, group, :, ::-1], gap - j - k - 1
                ).mean()
    return X_new


class SingularSpectrumAnalysis(BaseEstimator, UnivariateTransformerMixin):
    """Singular Spectrum Analysis.

    Parameters
    ----------
    window_size : int or float (default = 4)
        Size of the sliding window (i.e. the size of each word). If float, it
        represents the percentage of the size of each time series and must be
        between 0 and 1. The window size will be computed as
        ``max(2, ceil(window_size * n_timestamps))``.

    groups : None, int or array-like (default = None)
        The way the elementary matrices are grouped. If None, no grouping is
        performed. If an integer, it represents the number of groups and the
        bounds of the groups are computed as
        ``np.linspace(0, window_size, groups + 1).astype('int64')``.
        If array-like, each element must be array-like and contain the indices
        for each group.

    References
    ----------
    .. [1] N. Golyandina, and A. Zhigljavsky, "Singular Spectrum Analysis for
           Time Series". Springer-Verlag Berlin Heidelberg (2013).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.decomposition import SingularSpectrumAnalysis
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = SingularSpectrumAnalysis(window_size=5)
    >>> X_new = transformer.transform(X)
    >>> X_new.shape
    (50, 5, 150)

    """

    def __init__(self, window_size=4, groups=None):
        self.window_size = window_size
        self.groups = groups

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_splits, n_timestamps)
            Transformed data. ``n_splits`` value depends on the value of
            ``groups``. If ``groups=None``, ``n_splits`` is equal to
            ``window_size``. If ``groups`` is an integer, ``n_splits`` is
            equal to ``groups``. If ``groups`` is array-like, ``n_splits``
            is equal to the length of ``groups``. If ``n_split=1``, ``X_new``
            is squeezed and its shape is (n_samples, n_timestamps).

        """
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape
        window_size = self._check_params(n_timestamps)
        n_windows = n_timestamps - window_size + 1

        X_window = np.transpose(
            _windowed_view(X, n_samples, n_timestamps,
                           window_size, window_step=1), axes=(0, 2, 1)
        ).copy()
        X_tranpose = np.matmul(X_window,
                               np.transpose(X_window, axes=(0, 2, 1)))
        w, v = np.linalg.eigh(X_tranpose)
        w, v = w[:, ::-1], v[:, :, ::-1]

        X_elem = _outer_dot(v, X_window, n_samples, window_size, n_windows)
        X_groups, grouping_size = self._grouping(
            X_elem, n_samples, window_size, n_windows
        )
        if window_size >= n_windows:
            X_groups = np.transpose(X_groups, axes=(0, 1, 3, 2))
            gap = window_size
        else:
            gap = n_windows

        X_ssa = _diagonal_averaging(
            X_groups, n_samples, n_timestamps, window_size,
            n_windows, grouping_size, gap
        )
        return np.squeeze(X_ssa)

    def _grouping(self, X, n_samples, window_size, n_windows):
        if self.groups is None:
            grouping_size = window_size
            X_new = X
        elif isinstance(self.groups, int):
            grouping = np.linspace(0, window_size,
                                   self.groups + 1).astype('int64')
            grouping_size = len(grouping) - 1
            X_new = np.zeros((n_samples, grouping_size,
                              window_size, n_windows))
            for i, (j, k) in enumerate(zip(grouping[:-1], grouping[1:])):
                X_new[:, i] = X[:, j:k].sum(axis=1)
        else:
            grouping_size = len(self.groups)
            X_new = np.zeros((n_samples, grouping_size,
                              window_size, n_windows))
            for i, group in enumerate(self.groups):
                X_new[:, i] = X[:, group].sum(axis=1)
        return X_new, grouping_size

    def _check_params(self, n_timestamps):
        if not isinstance(self.window_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_size' must be an integer or a float.")
        if isinstance(self.window_size, (int, np.integer)):
            if not 2 <= self.window_size <= n_timestamps:
                raise ValueError(
                    "If 'window_size' is an integer, it must be greater "
                    "than or equal to 2 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.window_size)
                )
            window_size = self.window_size
        else:
            if not 0 < self.window_size <= 1:
                raise ValueError(
                    "If 'window_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 "
                    "(got {0}).".format(self.window_size)
                )
            window_size = max(2, ceil(self.window_size * n_timestamps))
        if not (self.groups is None
                or isinstance(self.groups, (int, list, tuple, np.ndarray))):
            raise TypeError("'groups' must be either None, an integer "
                            "or array-like.")
        if isinstance(self.groups, (int, np.integer)):
            if not 1 <= self.groups <= self.window_size:
                raise ValueError(
                    "If 'groups' is an integer, it must be greater than or "
                    "equal to 1 and lower than or equal to 'window_size'."
                )
        if isinstance(self.groups, (list, tuple, np.ndarray)):
            idx = np.concatenate(self.groups)
            diff = np.setdiff1d(idx, np.arange(self.window_size))
            flat_list = [item for group in self.groups for item in group]
            if ((diff.size > 0)
                or not (all(isinstance(x, (int, np.integer))
                            for x in flat_list))):
                raise ValueError(
                    "If 'groups' is array-like, all the values in 'groups' "
                    "must be integers between 0 and ('window_size' - 1)."
                )
        return window_size
